# cython: language_level=3

import datetime
import struct
from collections import OrderedDict

from tdxpy.parser.base import BaseParser


class GetHistoryTransactionData(BaseParser):
    date = None

    def setParams(self, market, code, date, start, count):
        if type(code) is str:
            code = code.encode("utf-8")

        pkg = bytearray.fromhex("01 01 30 00 02 01 16 00 16 00 06 24")
        pkg.extend(struct.pack("<IB9siH", date, market, code, start, count))

        self.send_pkg = pkg
        self.date = date

    def parseResponse(self, body_buf):

        pos = 0
        market, code, _, num = struct.unpack("<B9s4sH", body_buf[pos: pos + 16])

        pos += 16
        result = []

        for _ in range(num):

            raw_time, price, volume, zengcang, direction = struct.unpack("<HIIiH", body_buf[pos: pos + 16])

            pos += 16
            year = self.date // 10000
            month = self.date % 10000 // 100
            day = self.date % 100
            hour = raw_time // 60
            minute = raw_time % 60
            second = direction % 10000
            nature = direction  # 为了老用户接口的兼容性，已经转换为使用 nature_value
            value = direction // 10000
            nature_name = "换手"

            # 对于大于59秒的值，属于无效数值
            if second > 59:
                second = 0

            date = datetime.datetime(year, month, day, hour, minute, second)

            if value == 0:
                direction = 1
                if zengcang > 0:
                    if volume > zengcang:
                        nature_name = "多开"
                    elif volume == zengcang:
                        nature_name = "双开"
                elif zengcang == 0:
                    nature_name = "多换"
                else:
                    if volume == -zengcang:
                        nature_name = "双平"
                    else:
                        nature_name = "空平"
            elif value == 1:
                direction = -1
                if zengcang > 0:
                    if volume > zengcang:
                        nature_name = "空开"
                    elif volume == zengcang:
                        nature_name = "双开"
                elif zengcang == 0:
                    nature_name = "空换"
                else:
                    if volume == -zengcang:
                        nature_name = "双平"
                    else:
                        nature_name = "多平"
            else:
                direction = 0
                if zengcang > 0:
                    if volume > zengcang:
                        nature_name = "开仓"
                    elif volume == zengcang:
                        nature_name = "双开"
                elif zengcang < 0:
                    if volume > -zengcang:
                        nature_name = "平仓"
                    elif volume == -zengcang:
                        nature_name = "双平"
                else:
                    nature_name = "换手"

            if market in [31, 48]:
                if nature == 0:
                    direction = 1
                    nature_name = "B"
                elif nature == 256:
                    direction = -1
                    nature_name = "S"
                else:  # 512
                    direction = 0
                    nature_name = ""

            result.append(
                OrderedDict(
                    [
                        ("date", date),
                        ("hour", hour),
                        ("minute", minute),
                        ("price", price),
                        ("volume", volume),
                        ("zengcang", zengcang),
                        ("natrue_name", nature_name),
                        ("nature_name", nature_name),  # 修正了nature_name的拼写错误(natrue), 为了保持兼容性，原有的natrue_name还会保留一段时间
                        ("direction", direction),
                        ("nature", nature),
                    ]
                )
            )

        return result



from .block_reader import BlockReader
from .block_reader import CustomerBlockReader
from .daily_bar_reader import TdxDailyBarReader
from .daily_bar_reader import TdxFileNotFoundException
from .daily_bar_reader import TdxNotAssignVipdocPathException
from .exhq_daily_bar_reader import TdxExHqDailyBarReader
from .history_financial_reader import HistoryFinancialReader
from .lc_min_bar_reader import TdxLCMinBarReader
from .min_bar_reader import TdxMinBarReader

__all__ = [
    "TdxNotAssignVipdocPathException",
    "TdxFileNotFoundException",
    "HistoryFinancialReader",
    "TdxExHqDailyBarReader",
    "CustomerBlockReader",
    "TdxLCMinBarReader",
    "TdxDailyBarReader",
    "TdxMinBarReader",
    "BlockReader",
    # "GBBQReader",
]



# cython: language_level=3

import struct
from collections import OrderedDict

from tdxpy.parser.base import BaseParser


class GetMarkets(BaseParser):
    def setup(self):
        self.send_pkg = bytearray.fromhex("01 02 48 69 00 01 02 00 02 00 f4 23")

    def parseResponse(self, body_buf):
        """

        :param body_buf:
        :return:
        """
        pos = 0
        cnt, = struct.unpack("<H", body_buf[pos: pos + 2])

        pos += 2
        result = []

        for _ in range(cnt):
            category, raw_name, market, raw_short_name, _, unknown_bytes = struct.unpack("<B32sB2s26s2s", body_buf[pos: pos + 64])
            pos += 64

            if category == 0 and market == 0:
                continue

            name = raw_name.decode("gbk")
            short_name = raw_short_name.decode("gbk")

            result.append(
                OrderedDict(
                    [
                        ("market", market),
                        ("category", category),
                        ("name", name.rstrip("\x00")),
                        ("short_name", short_name.rstrip("\x00")),
                        ('unknown_bytes', unknown_bytes)
                    ]
                )
            )

        return result



# cython: language_level=3

import struct

from tdxpy.parser.base import BaseParser


class GetSecurityCountCmd(BaseParser):
    def setParams(self, market):
        """

        :param market:
        """
        pkg = bytearray.fromhex("0c 0c 18 6c 00 01 08 00 08 00 4e 04")

        pkg.extend(struct.pack("<H", market))
        pkg.extend(b"\x75\xc7\x33\x01")

        self.send_pkg = pkg

    def parseResponse(self, body_buf):
        """

        :param body_buf:
        :return:
        """
        num, = struct.unpack("<H", body_buf[:2])

        return num



# cython: language_level=3

from pathlib import Path

import pandas as pd

from tdxpy.logger import logger
from tdxpy.reader.base_reader import BaseReader
from tdxpy.reader.base_reader import TdxFileNotFoundException
from tdxpy.reader.base_reader import TdxNotAssignVipdocPathException


class TdxDailyBarReader(BaseReader):
    """
    读取通达信日线数据
    """

    # 交易所
    SECURITY_EXCHANGE = ["sz", "sh"]

    # 股票类型
    SECURITY_TYPE = ["SH_A_STOCK", "SH_B_STOCK", "SH_INDEX", "SH_FUND", "SH_BOND", "SZ_A_STOCK", "SZ_B_STOCK", "SZ_INDEX", "SZ_FUND", "SZ_BOND"]

    # todo 股票系数
    SECURITY_COEFFICIENT = {
        "SH_A_STOCK": [0.01, 0.01],
        "SH_B_STOCK": [0.001, 0.01],
        "SH_INDEX": [0.01, 1.0],
        "SH_FUND": [0.001, 1.0],
        "SH_BOND": [0.001, 1.0],
        "SZ_A_STOCK": [0.01, 0.01],
        "SZ_B_STOCK": [0.01, 0.01],
        "SZ_INDEX": [0.01, 1.0],
        "SZ_FUND": [0.001, 0.01],
        "SZ_BOND": [0.001, 0.01],
    }

    def generate_filename(self, code, exchange):
        """

        :param code:
        :param exchange:
        :return:
        """
        if not self.vipdoc_path:
            raise TdxNotAssignVipdocPathException(r"Please provide a vipdoc path , such as c:\\newtdx\\vipdoc")  # noqa

        filename = Path(self.vipdoc_path, exchange, "lday", f"{exchange}{code}.day")  # noqa

        return str(filename)

    def get_kline_by_code(self, code, exchange):
        """

        :param code:
        :param exchange:
        :return:
        """
        filename = self.generate_filename(code, exchange)

        return self.parse_data_by_file(filename)

    def parse_data_by_file(self, filename):
        """

        :param filename:
        :return:
        """
        if not Path(filename).is_file():
            raise TdxFileNotFoundException("no tdx kline data, please check path %s", filename)

        content = Path(filename).read_bytes()
        content = self.unpack_records("<IIIIIfII", content)

        return content

    def get_df(self, code_or_file, exchange=None):
        """

        :param code_or_file:
        :param exchange:
        :return:
        """
        if not exchange:
            return self.get_df_by_file(code_or_file)

        return self.get_df_by_code(code_or_file, exchange)

    def get_df_by_file(self, filename):
        """
        通过文件名读数据
        :param filename:
        :return:
        """
        if not Path(filename).is_file():
            raise TdxFileNotFoundException("no tdx kline data, please check path %s", filename)

        security_type = self.get_security_type(filename)

        if security_type not in self.SECURITY_TYPE:
            logger.exception("Unknown security type !\n")
            raise NotImplementedError

        coefficient = self.SECURITY_COEFFICIENT[security_type]
        data = [self._df_convert(row_, coefficient) for row_ in self.parse_data_by_file(filename)]

        df = pd.DataFrame(data=data, columns=["date", "open", "high", "low", "close", "amount", "volume"])
        df.index = pd.to_datetime(df.date, errors='coerce')
        # df.index = pd.to_datetime(df.date)

        return df[["open", "high", "low", "close", "amount", "volume"]]

    def get_df_by_code(self, code, exchange):
        """
        通过股票代码生成文件名
        :param code: 股票代码
        :param exchange: 交易所
        :return:
        """
        name = self.generate_filename(code, exchange)

        return self.get_df_by_file(name)

    @staticmethod
    def _df_convert(row_, coefficient):
        """
        源数据转换

        :param row_:
        :param coefficient:
        :return:
        """
        t_date = str(row_[0])
        datestr = t_date[:4] + "-" + t_date[4:6] + "-" + t_date[6:]

        new_row = (
            datestr,
            row_[1] * coefficient[0],  # * 0.01 * 1000 , zipline need 1000 times to original price
            row_[2] * coefficient[0],
            row_[3] * coefficient[0],
            row_[4] * coefficient[0],
            row_[5],
            row_[6] * coefficient[1],
        )

        return new_row

    def get_security_type(self, filename):
        """
        获取股票类型, A股, B股, 指数等

        :param filename:
        :return:
        """
        logger.debug(f'get_security_type name: {filename}')
        basename = Path(filename).stem

        # exchange = str(name[-12:-10]).lower()
        # code_head = name[-10:-8]

        exchange = str(basename[:2]).lower()
        code_head = basename[2:4]

        if exchange == self.SECURITY_EXCHANGE[0]:
            if code_head in ["00", "30"]:
                return "SZ_A_STOCK"

            if code_head in ["20"]:
                return "SZ_B_STOCK"

            if code_head in ["39"]:
                return "SZ_INDEX"

            if code_head in ["15", "16"]:
                return "SZ_FUND"

            if code_head in ["10", "11", "12", "13", "14"]:
                return "SZ_BOND"

        if exchange == self.SECURITY_EXCHANGE[1]:
            if code_head in ["60", "68"]:  # 688XXX科创板
                return "SH_A_STOCK"

            if code_head in ["90"]:
                return "SH_B_STOCK"

            if code_head in ["00", "88", "99"]:
                return "SH_INDEX"

            if code_head in ["50", "51"]:
                return "SH_FUND"

            if code_head in ["01", "10", "11", "12", "13", "14", "20"]:
                return "SH_BOND"

        logger.debug("Unknown security exchange !")
        raise NotImplementedError



# cython: language_level=3

import struct
from collections import OrderedDict

from tdxpy.helper import get_volume
from tdxpy.parser.base import BaseParser


class GetSecurityList(BaseParser):
    def setParams(self, market, start):
        """
        设置参数
        :param market: 市场
        :param start: 开始位置
        """
        pkg = bytearray.fromhex("0c 01 18 64 01 01 06 00 06 00 50 04")
        pkg.extend(struct.pack("<HH", market, start))

        self.send_pkg = pkg

    def parseResponse(self, body_buf):
        pos = 0
        num, = struct.unpack("<H", body_buf[:2])

        pos += 2
        symbols = []

        for _ in range(num):
            one_bytes = body_buf[pos: pos + 29]

            code, volunit, name_bytes, reversed_bytes1, decimal_point, pre_close_raw, reversed_bytes2 = struct.unpack("<6sH8s4sBI4s", one_bytes)

            code = code.decode("utf-8", errors='ignore')
            name = name_bytes.decode('gbk', errors='ignore')

            pre_close = get_volume(pre_close_raw)
            pos += 29

            rows = OrderedDict(
                [
                    ("code", code),
                    ("volunit", volunit),
                    ("decimal_point", decimal_point),
                    ("name", name),
                    ("pre_close", pre_close),
                ]
            )

            symbols.append(rows)

        return symbols



# cython: language_level=3

import struct
from collections import OrderedDict

from tdxpy.helper import get_price
from tdxpy.parser.base import BaseParser


class GetMinuteTimeData(BaseParser):
    def setParams(self, market, code):
        """
        设置参数
        :param market: 市场
        :param code: 代码
        """
        if type(code) is str:
            code = code.encode("utf-8")

        pkg = bytearray.fromhex("0c 1b 08 00 01 01 0e 00 0e 00 1d 05")
        pkg.extend(struct.pack("<H6sI", market, code, 0))

        self.send_pkg = pkg

    def parseResponse(self, body_buf):
        pos = 0

        num, = struct.unpack("<H", body_buf[:2])

        last_price = 0

        pos += 4
        prices = []

        for _ in range(num):
            price_raw, pos = get_price(body_buf, pos)
            reversed1, pos = get_price(body_buf, pos)

            vol, pos = get_price(body_buf, pos)
            last_price = float(last_price + price_raw)

            prices.append(OrderedDict([("price", last_price / 100), ("vol", vol)]))

        return prices



# cython: language_level=3

import abc
import tempfile
from urllib.request import Request
from urllib.request import urlopen

from tdxpy.logger import logger


def fetch_report_hook(downloaded, total_size):
    logger.debug(f"Downloaded {downloaded}, Total is {total_size}")


class BaseCrawler:
    mode = "http"

    def fetch_and_parse(self, reporthook=None, path_to_download=None, proxies=None, chunksize=1024 * 50, *args, **kwargs):
        """
        function to get data ,
        :param chunksize:
        :param args:
        :param kwargs:
        :return:
        :param reporthook 使用urllib.request 的report_hook 来汇报下载进度 \
                    参考 https://docs.python.org/3/library/urllib.request.html#module-urllib.request
        :param path_to_download 数据文件下载的地址，如果没有提供，则下载到临时文件中，并在解析之后删除
        :param proxies urllib格式的代理服务器设置
        :return: 解析之后的数据结果
        """

        method = ("get_content", "fetch_via_http")[self.mode == "http"]
        download_file = getattr(self, method)(reporthook=reporthook, path_to_download=path_to_download, proxies=proxies, chunksize=chunksize, *args, **kwargs)
        result = self.parse(download_file, *args, **kwargs)
        download_file.close()

        return result

    def fetch_via_http(self, reporthook=None, path_to_download=None, chunksize=1024 * 50, *args, **kwargs):
        """

        :param reporthook:
        :param path_to_download:
        :param chunksize:
        :param args:
        :param kwargs:
        :return:
        """
        download_file = path_to_download and open(path_to_download, "wb") or tempfile.NamedTemporaryFile(delete=True)
        url = self.get_url(*args, **kwargs)

        request = Request(url)
        request.add_header("Referer", url)
        request.add_header("User-Agent", r"Mozilla/5.0 (Macintosh Intel Mac OS X 10_14_0) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/76.0.3809.100 Safari/537.36")
        response = urlopen(request)

        if response.getheader("Content-Length") is not None:
            total_size = int(response.getheader("Content-Length").strip())
            downloaded = 0

            while True:
                chunk = response.read(chunksize)
                downloaded += len(chunk)

                reporthook and reporthook(downloaded, total_size)

                if not chunk:
                    break

                download_file.write(chunk)
        else:
            content = response.read()
            download_file.write(content)

        download_file.seek(0)
        return download_file

    @abc.abstractmethod
    def get_url(self, *args, **kwargs):
        """

        :param args:
        :param kwargs:
        """
        raise NotImplementedError("will impl in subclass")

    @abc.abstractmethod
    def get_content(self, reporthook=None, path_to_download=None, proxies=None, chunksize=1024 * 50, *args, **kwargs):
        """

        :param reporthook:
        :param path_to_download:
        :param proxies:
        :param chunksize:
        :param args:
        :param kwargs:
        """
        raise NotImplementedError("will impl in subclass")

    @abc.abstractmethod
    def parse(self, download_file, *args, **kwargs):
        """

        :param download_file:
        :param args:
        :param kwargs:
        """
        raise NotImplementedError("will impl in subclass")



# cython: language_level=3

import struct
from collections import OrderedDict

from tdxpy.helper import get_price
from tdxpy.helper import get_time
from tdxpy.parser.base import BaseParser


class GetHistoryTransactionData(BaseParser):
    def setParams(self, market, code, start, count, date):
        """

        :param market:
        :param code:
        :param start:
        :param count:
        :param date:
        """
        if type(code) is str:
            code = code.encode("utf-8")

        if type(date) is (type(date) is str) or (type(date) is bytes):
            date = int(date)

        pkg = bytearray.fromhex("0c 01 30 01 00 01 12 00 12 00 b5 0f")
        pkg.extend(struct.pack("<IH6sHH", date, market, code, start, count))

        self.send_pkg = pkg

    def parseResponse(self, body_buf):
        """

        :param body_buf:
        :return:
        """
        pos = 0
        num, = struct.unpack("<H", body_buf[:2])

        pos += 2
        ticks = []

        # skip 4 bytes
        pos += 4
        last_price = 0

        for _ in range(num):
            hour, minute, pos = get_time(body_buf, pos)  # noqa

            price_raw, pos = get_price(body_buf, pos)
            vol, pos = get_price(body_buf, pos)

            buy_or_sell, pos = get_price(body_buf, pos)
            _, pos = get_price(body_buf, pos)

            last_price = last_price + price_raw

            tick = OrderedDict(
                [
                    ("time", "%02d:%02d" % (hour, minute)),
                    ("price", float(last_price) / 100),
                    ("vol", vol),
                    ("buyorsell", buy_or_sell),
                ]
            )

            ticks.append(tick)

        return ticks



# cython: language_level=3

import time
from threading import Thread

from tdxpy.logger import logger

# 10秒一个heartbeat
DEFAULT_HEARTBEAT_INTERVAL = 10.0


class HeartBeatThread(Thread):
    # 参考 :https://stackoverflow.com/questions/6524459/stopping-a-thread-after-a-certain-amount-of-time

    def __init__(self, client, stop_event, interval=DEFAULT_HEARTBEAT_INTERVAL):
        """
        初始化
        :param client: 客户端主进程
        :param stop_event: 停止事件进程
        :param interval: 心跳周期
        """

        self.heartbeat_interval = interval
        self.stop_event = stop_event

        self.client = client.client
        self.api = client

        super().__init__()

    def run(self):
        # stop_event 未设置就执行
        while not self.stop_event.is_set():
            # 客户端主进程 and 满足 heartbeat_interval
            if self.client and (time.time() - self.api.last_ack_time > self.heartbeat_interval):
                try:
                    # 发送一个获取股票数量的包作为心跳包
                    logger.debug('发送一个获取股票数量的包作为心跳包...')
                    self.api.do_heartbeat()
                except Exception as e:
                    logger.exception(str(e))

            # 等待心跳周期
            self.stop_event.wait(self.heartbeat_interval)



# cython: language_level=3

import struct
from collections import OrderedDict

from tdxpy.helper import get_datetime
from tdxpy.helper import get_volume
from tdxpy.parser.base import BaseParser

XDXR_CATEGORY_MAPPING = {
    1: "除权除息",
    2: "送配股上市",
    3: "非流通股上市",
    4: "未知股本变动",
    5: "股本变化",
    6: "增发新股",
    7: "股份回购",
    8: "增发新股上市",
    9: "转配股上市",
    10: "可转债上市",
    11: "扩缩股",
    12: "非流通股缩股",
    13: "送认购权证",
    14: "送认沽权证",
}


class GetXdXrInfo(BaseParser):
    def setParams(self, market, code):
        """
        设置参数
        :param market: 市场
        :param code: 股票代码
        """
        if type(code) is str:
            code = code.encode("utf-8")

        pkg = bytearray.fromhex("0c 1f 18 76 00 01 0b 00 0b 00 0f 00 01 00")
        pkg.extend(struct.pack("<B6s", market, code))

        self.send_pkg = pkg

    def parseResponse(self, body_buf):
        """
        解析返回结果
        :param body_buf:
        :return:
        """
        pos = 0

        if len(body_buf) < 11:
            return []

        pos += 9  # skip 9
        num, = struct.unpack("<H", body_buf[pos: pos + 2])

        pos += 2
        rows = []

        def _get_v(v):
            if v == 0:
                return 0

            return get_volume(v)

        for _ in range(num):
            pos += 7
            pos += 1

            year, month, day, hour, minute, pos = get_datetime(9, body_buf, pos)
            category, = struct.unpack("<B", body_buf[pos: pos + 1])
            pos += 1

            suogu = None
            panqianliutong, panhouliutong, qianzongguben, houzongguben = None, None, None, None
            songzhuangu, fenhong, peigu, peigujia = None, None, None, None
            fenshu, xingquanjia = None, None

            if category == 1:
                fenhong, peigujia, songzhuangu, peigu = struct.unpack("<ffff", body_buf[pos: pos + 16])
            elif category in [11, 12]:
                _, _, suogu, _ = struct.unpack("<IIfI", body_buf[pos: pos + 16])
            elif category in [13, 14]:
                xingquanjia, _, fenshu, _ = struct.unpack("<fIfI", body_buf[pos: pos + 16])
            else:
                panqianliutong_raw, qianzongguben_raw, panhouliutong_raw, houzongguben_raw = struct.unpack("<IIII", body_buf[pos: pos + 16])
                panqianliutong = _get_v(panqianliutong_raw)
                panhouliutong = _get_v(panhouliutong_raw)
                qianzongguben = _get_v(qianzongguben_raw)
                houzongguben = _get_v(houzongguben_raw)

            pos += 16

            row = OrderedDict(
                [
                    ("year", year),
                    ("month", month),
                    ("day", day),
                    ("category", category),
                    ("name", self.get_category_name(category)),
                    ("fenhong", fenhong),
                    ("peigujia", peigujia),
                    ("songzhuangu", songzhuangu),
                    ("peigu", peigu),
                    ("suogu", suogu),
                    ("panqianliutong", panqianliutong),
                    ("panhouliutong", panhouliutong),
                    ("qianzongguben", qianzongguben),
                    ("houzongguben", houzongguben),
                    ("fenshu", fenshu),
                    ("xingquanjia", xingquanjia),
                ]
            )
            rows.append(row)

        return rows

    @staticmethod
    def get_category_name(category_id):
        """
        获取分类名称
        如果配置内没有，则返回 category_id
        :param category_id:
        :return: mixed
        """
        return XDXR_CATEGORY_MAPPING.get(category_id, str(category_id))



# cython: language_level=3

import struct

from tdxpy.parser.base import BaseParser


class GetReportFile(BaseParser):
    def setParams(self, filename, offset=0):
        """
        设置参数
        :param filename: 文件名
        :param offset: 获取数量
        """
        pkg = bytearray.fromhex("0C 12 34 00 00 00")

        # Fom DTGear request.py file
        node_size = 0x7530

        raw_data = struct.pack(r"<H2I100s", 0x06B9, offset, node_size, filename.encode("utf-8"))
        raw_data_len = struct.calcsize(r"<H2I100s")

        pkg.extend(struct.pack("<HH{}s".format(raw_data_len), raw_data_len, raw_data_len, raw_data))

        self.send_pkg = pkg

    def parseResponse(self, body_buf):
        """

        :param body_buf:
        :return:
        """

        chunk_size, = struct.unpack("<I", body_buf[:4])

        if chunk_size > 0:
            return {"chunksize": chunk_size, "chunkdata": body_buf[4:]}

        return {"chunksize": 0}



# cython: language_level=3

import struct

from tdxpy import logger
from tdxpy.parser.base import BaseParser
from tdxpy.reader.block_reader import BlockReader
from tdxpy.reader.block_reader import BlockReader_TYPE_FLAT


class GetBlockInfoMeta(BaseParser):
    def setParams(self, block_file):
        """
        设置参数
        :param block_file:
        """
        if type(block_file) is str:
            block_file = block_file.encode("utf-8")

        pkg = bytearray.fromhex("0C 39 18 69 00 01 2A 00 2A 00 C5 02")
        pkg.extend(struct.pack("<{}s".format(0x2A - 2), block_file))

        self.send_pkg = pkg

    def parseResponse(self, body_buf):
        """
        解析结果
        :param body_buf:
        :return:
        """
        size, _, hash_value, _ = struct.unpack("<I1s32s1s", body_buf)

        return {"size": size, "hash_value": hash_value}


class GetBlockInfo(BaseParser):
    def setParams(self, block_file, start, size):
        """
        设置参数
        :param block_file: 板块文件名
        :param start: 开始位置
        :param size: 容量
        """
        if type(block_file) is str:
            block_file = block_file.encode("utf-8")

        pkg = bytearray.fromhex("0c 37 18 6a 00 01 6e 00 6e 00 b9 06")
        pkg.extend(struct.pack("<II{}s".format(0x6E - 10), start, size, block_file))

        self.send_pkg = pkg

    def parseResponse(self, body_buf):
        """
        解析结果
        :param body_buf:
        :return:
        """
        return body_buf[4:]


def get_and_parse_block_info(client, block_file):
    """
    获取价格和板块信息
    :param client: 客户端
    :param block_file: 板块文件名
    :return:
    """
    try:
        meta = client.get_block_info_meta(block_file)

        if not meta:
            return None
    except Exception as exc:
        logger.logger.exception(exc)
        return None

    size = meta["size"]
    one_chunk = 0x7530

    chunks = size // one_chunk

    if size % one_chunk != 0:
        chunks += 1

    file_content = bytearray()

    for seg in range(chunks):
        piece_data = client.get_block_info(block_file, seg * one_chunk, size)
        file_content.extend(piece_data)

    content = BlockReader().get_data(file_content, BlockReader_TYPE_FLAT)

    return content



# cython: language_level=3

from tdxpy.parser.base import BaseParser


class BaseSetup(BaseParser):
    def parseResponse(self, body_buf):
        """
        解析返回结果
        :param body_buf:
        :return:
        """
        return body_buf


class SetupCmd1(BaseSetup):
    def setup(self):
        self.send_pkg = bytearray.fromhex(
            "0c 02 18 93 00 01 03 00 03 00 0d 00 01"
        )


class SetupCmd2(BaseSetup):
    def setup(self):
        self.send_pkg = bytearray.fromhex(
            "0c 02 18 94 00 01 03 00 03 00 0d 00 02"
        )


class SetupCmd3(BaseSetup):
    def setup(self):
        self.send_pkg = bytearray.fromhex(
            "0c 03 18 99 00 01 20 00 20 00 db 0f d5 d0"
            "c9 cc d6 a4 a8 af 00 00 00 8f c2 25 40 13"
            "00 00 d5 00 c9 cc bd f0 d7 ea 00 00 00 02"
        )


class ExSetupCmd1(BaseParser):  # noqa
    def setup(self):
        self.send_pkg = bytearray.fromhex(
            "01 01 48 65 00 01 52 00 52 00 54 24 1f 32 c6 e5"
            "d5 3d fb 41 1f 32 c6 e5 d5 3d fb 41 1f 32 c6 e5"
            "d5 3d fb 41 1f 32 c6 e5 d5 3d fb 41 1f 32 c6 e5"
            "d5 3d fb 41 1f 32 c6 e5 d5 3d fb 41 1f 32 c6 e5"
            "d5 3d fb 41 1f 32 c6 e5 d5 3d fb 41 cc e1 6d ff"
            "d5 ba 3f b8 cb c5 7a 05 4f 77 48 ea"
        )



# cython: language_level=3

import struct
from collections import OrderedDict
from pathlib import Path

import pandas as pd

from tdxpy.reader.base_reader import BaseReader

"""
参考这个 http://blog.csdn.net/Metal1/article/details/44352639
"""

BlockReader_TYPE_FLAT = 0
BlockReader_TYPE_GROUP = 1


class BlockReader(BaseReader):
    def get_df(self, name, result_type=BlockReader_TYPE_FLAT):
        """
        转换 pd.DataFrame
        :param name:
        :param result_type:
        :return:
        """

        result = self.get_data(name, result_type)

        return pd.DataFrame(result)

    @staticmethod
    def get_data(name, result_type=BlockReader_TYPE_FLAT):
        """
        解析数据
        :param name:
        :param result_type:
        :return:
        """
        result = []

        data = (type(name) is bytearray) and name or Path(name).read_bytes()

        pos = 384

        num, = struct.unpack("<H", data[pos: pos + 2])

        pos += 2

        for i in range(num):
            block_name_raw = data[pos: pos + 9]

            pos += 9

            block_name = block_name_raw.decode("gbk", "ignore").rstrip("\x00")
            stock_count, block_type = struct.unpack("<HH", data[pos: pos + 4])

            pos += 4

            block_stock_begin = pos
            codes = []

            for code_index in range(stock_count):
                one_code = data[pos: pos + 7].decode("utf-8", "ignore").rstrip("\x00")

                pos += 7

                if result_type == BlockReader_TYPE_FLAT:
                    result.append(
                        OrderedDict(
                            [
                                ("blockname", block_name),  # noqa
                                ("block_type", block_type),
                                ("code_index", code_index),
                                ("code", one_code),
                            ]
                        )
                    )
                elif result_type == BlockReader_TYPE_GROUP:
                    codes.append(one_code)

            if result_type == BlockReader_TYPE_GROUP:
                result.append(OrderedDict([("blockname", block_name), ("block_type", block_type), ("stock_count", stock_count), ("code_list", ",".join(codes))]))  # noqa

            pos = block_stock_begin + 2800

        return result


class CustomerBlockReader(BaseReader):
    """
    读取通达信备份的自定义板块文件夹，返回格式与通达信板块一致，在广发证券客户端上测试通过，其它未测试
    """

    def get_df(self, name, result_type=BlockReader_TYPE_FLAT):
        """
        转换 pd.DataFrame
        :param name:
        :param result_type:
        :return:
        """
        result = self.get_data(name, result_type)
        return pd.DataFrame(result)

    @staticmethod
    def get_data(name, result_type=BlockReader_TYPE_FLAT):
        """
        解析数据
        :param name:
        :param result_type:
        :return:
        """
        if not Path(name).is_dir():
            raise Exception("not a directory")

        block_file = "/".join([name, "blocknew.cfg"])  # noqa

        if not Path(block_file).exists():
            raise Exception("file not exists")

        block_data = open(block_file, "rb").read()

        pos = 0
        result = []

        while pos < len(block_data):
            n1 = block_data[pos: pos + 50].decode("gbk", "ignore").rstrip("\x00")
            n2 = block_data[pos + 50: pos + 120].decode("gbk", "ignore").rstrip("\x00")

            pos = pos + 120

            n1 = n1.split("\x00")[0]
            n2 = n2.split("\x00")[0]
            bf = "/".join([name, n2 + ".blk"])

            if not Path(bf).exists():
                raise Exception("file not exists")

            codes = open(bf).read().splitlines()

            if result_type == BlockReader_TYPE_FLAT:
                for index, code in enumerate(codes):
                    code and result.append(OrderedDict([("blockname", n1), ("block_type", n2), ("code_index", index), ("code", code[1:])]))  # noqa

            if result_type == BlockReader_TYPE_GROUP:
                cc = [c[1:] for c in codes if c != ""]
                result.append(OrderedDict([("blockname", n1), ("block_type", n2), ("stock_count", len(cc)), ("code_list", ",".join(cc))]))  # noqa

        return result



# cython: language_level=3

from collections import OrderedDict
from pathlib import Path

import pandas as pd

from tdxpy.reader.base_reader import BaseReader
from tdxpy.reader.base_reader import TdxFileNotFoundException

"""
网传秘籍...

二、通达信5分钟线*.lc5文件和*.lc1文件 文件名即股票代码 每32个字节为一个5分钟数据，每字段内低字节在前 00 ~ 01 字节：日期，整型，
设其值为num，则日期计算方法为： year=floor(num/2048)+2004; month=floor(mod(num,2048)/100); day=mod(mod(num,2048),100);
02 ~ 03 字节： 从0点开始至目前的分钟数，整型 04 ~ 07 字节：开盘价，float型 08 ~ 11 字节：最高价，float型 12 ~ 15 字节：最低价，
float型 16 ~ 19 字节：收盘价，float型 20 ~ 23 字节：成交额，float型 24 ~ 27 字节：成交量（股），整型 28 ~ 31 字节：（保留）

"""


class TdxLCMinBarReader(BaseReader):
    """
    读取通达信分钟数据
    """

    def parse_data_by_file(self, filename):
        """

        :param filename:
        :return:
        """
        if not Path(filename).is_file():
            raise TdxFileNotFoundException(f"no tdx kline data, please check path {filename}")

        content = Path(filename).read_bytes()
        raw_li = self.unpack_records("<HHfffffII", content)  # noqa

        data = []

        for row in raw_li:
            year, month, day = self._parse_date(row[0])
            hour, minute = self._parse_time(row[1])

            data.append(
                OrderedDict(
                    [
                        ("date", f"{year:04d}-{month:02d}-{day:02d} {hour:02d}:{minute:02d}"),
                        ("year", year),
                        ("month", month),
                        ("day", day),
                        ("hour", hour),
                        ("minute", minute),
                        ("open", row[2]),
                        ("high", row[3]),
                        ("low", row[4]),
                        ("close", row[5]),
                        ("amount", row[6]),
                        ("volume", row[7]),
                        # ('unknown', row[8])
                    ]
                )
            )

        return data

    def get_df(self, code_or_file, **kwargs):
        """
        :param code_or_file:
        :param kwargs:
        :return:
        """
        df = pd.DataFrame(data=self.parse_data_by_file(code_or_file))
        df.index = pd.to_datetime(df.date)
        df = df[["open", "high", "low", "close", "amount", "volume"]]

        return df



# cython: language_level=3

class TdxConnectionError(Exception):
    """
    当连接服务器出错的时候，会抛出的异常
    """

    pass


class TdxFunctionCallError(Exception):
    """
    当行数调用出错的时候
    """

    def __init__(self, *args):
        super(TdxFunctionCallError, self).__init__(*args)
        self.original_exception = None


class ValidationException(Exception):
    ...



# cython: language_level=3

import struct
from pathlib import Path

import pandas as pd

from tdxpy.reader.base_reader import BaseReader
from tdxpy.reader.base_reader import TdxFileNotFoundException


class TdxExHqDailyBarReader(BaseReader):
    """
    读取通达信数据
    """

    def parse_data_by_file(self, filename):
        """
        按文件名解析内容
        :param filename:
        :return:
        """
        if not Path(filename).is_file():
            raise TdxFileNotFoundException(f"no tdx kline data, please check path {filename}")

        content = Path(filename).read_bytes()
        content = self.unpack_records("<IffffIIf", content)  # noqa

        return content

    def get_df(self, code_or_file, **kwargs):
        """
        转换 pd.DataFrame 格式
        :param code_or_file:
        :param kwargs:
        :return:
        """
        columns = ["date", "open", "high", "low", "close", "amount", "volume", "jiesuan", "hk_stock_amount"]  # noqa
        data = [self._df_convert(row) for row in self.parse_data_by_file(code_or_file)]

        df = pd.DataFrame(data=data, columns=columns)
        df.index = pd.to_datetime(df.date)
        df = df[["open", "high", "low", "close", "amount", "volume", "jiesuan", "hk_stock_amount"]]  # noqa

        return df

    @staticmethod
    def _df_convert(row):
        """

        :param row:
        :return:
        """
        t_date = str(row[0])
        datestr = t_date[:4] + "-" + t_date[4:6] + "-" + t_date[6:]

        hk_stock_amount, = struct.unpack("<f", struct.pack("<I", row[5]))
        new_row = (datestr, row[1], row[2], row[3], row[4], row[5], row[6], row[7], hk_stock_amount)

        return new_row



# cython: language_level=3

import struct

from tdxpy.parser.base import BaseParser


class GetCompanyInfoContent(BaseParser):
    def setParams(self, market, code, filename, start, length):
        """

        :param market:
        :param code:
        :param filename:
        :param start:
        :param length:
        """
        if type(code) is str:
            code = code.encode("utf-8")

        if type(filename) is str:
            filename = filename.encode("utf-8")

        if len(filename) != 80:
            filename = filename.ljust(80, b"\x00")

        pkg = bytearray.fromhex("0c 07 10 9c 00 01 68 00 68 00 d0 02")
        pkg.extend(struct.pack("<H6sH80sIII", market, code, 0, filename, start, length, 0))

        self.send_pkg = pkg

    def parseResponse(self, body_buf):
        """

        :param body_buf:
        :return:
        """
        pos = 0

        _, length = struct.unpack("<10sH", body_buf[:12])

        pos += 12

        content = body_buf[pos: pos + length]
        content = content.decode("gbk")

        return content



# cython: language_level=3

import secrets

import pandas as pd

from tdxpy.base_socket_client import BaseSocketClient
from tdxpy.base_socket_client import last_ack_time
from tdxpy.parser.setup_commands import SetupCmd1
from tdxpy.parser.setup_commands import SetupCmd2
from tdxpy.parser.setup_commands import SetupCmd3
from tdxpy.parser.std.get_block_info import GetBlockInfo
from tdxpy.parser.std.get_block_info import GetBlockInfoMeta
from tdxpy.parser.std.get_block_info import get_and_parse_block_info
from tdxpy.parser.std.get_company_info_category import GetCompanyInfoCategory
from tdxpy.parser.std.get_company_info_content import GetCompanyInfoContent
from tdxpy.parser.std.get_finance_info import GetFinanceInfo
from tdxpy.parser.std.get_history_minute_time_data import GetHistoryMinuteTimeData
from tdxpy.parser.std.get_history_transaction_data import GetHistoryTransactionData
from tdxpy.parser.std.get_index_bars import GetIndexBarsCmd
from tdxpy.parser.std.get_minute_time_data import GetMinuteTimeData
from tdxpy.parser.std.get_report_file import GetReportFile
from tdxpy.parser.std.get_security_bars import GetSecurityBarsCmd
from tdxpy.parser.std.get_security_count import GetSecurityCountCmd
from tdxpy.parser.std.get_security_list import GetSecurityList
from tdxpy.parser.std.get_security_quotes import GetSecurityQuotesCmd
from tdxpy.parser.std.get_transaction_data import GetTransactionData
from tdxpy.parser.std.get_xdxr_info import GetXdXrInfo


class TdxHq_API(BaseSocketClient):  # noqa
    def setup(self):
        SetupCmd1(self.client).call_api()
        SetupCmd2(self.client).call_api()
        SetupCmd3(self.client).call_api()

    @last_ack_time
    def get_security_bars(self, category, market, code, start, count):
        """
        # Notice：，如果一个股票当天停牌，那天的K线还是能取到，成交量为0
        :param category:
        :param market:
        :param code:
        :param start:
        :param count:
        :return:
        """
        cmd = GetSecurityBarsCmd(self.client, lock=self.lock)
        cmd.setParams(category, market, code, start, count)

        return cmd.call_api()

    @last_ack_time
    def get_index_bars(self, category, market, code, start, count):
        """
        指数时间线
        :param category:
        :param market:
        :param code:
        :param start:
        :param count:
        :return:
        """
        cmd = GetIndexBarsCmd(self.client, lock=self.lock)
        cmd.setParams(category, market, code, start, count)

        return cmd.call_api()

    @last_ack_time
    def get_security_quotes(self, all_stock=None, code=None):
        """
        支持三种形式的参数
        get_security_quotes(market, code )
        get_security_quotes((market, code))
        get_security_quotes([(market1, code1), (market2, code2)] )

        :rtype: object
        :param all_stock （market, code) 的数组
        :param code{optional} code to query
        :return:
        """

        # if not all_stock:
        #     raise ValidationException('参数错误')

        if code:
            all_stock = [(all_stock, code)]
        elif (isinstance(all_stock, list) or isinstance(all_stock, tuple)) and len(all_stock) == 2 and type(all_stock[0]) is int:
            all_stock = [all_stock]

        cmd = GetSecurityQuotesCmd(self.client, lock=self.lock)
        cmd.setParams(all_stock)

        return cmd.call_api()

    @last_ack_time
    def get_security_count(self, market):
        """
        获取证券数量
        :param market:
        :return:
        """
        cmd = GetSecurityCountCmd(self.client, lock=self.lock)
        cmd.setParams(market)

        return cmd.call_api()

    @last_ack_time
    def get_security_list(self, market, start):
        """
        获取证券列表
        :param market:
        :param start:
        :return:
        """
        cmd = GetSecurityList(self.client, lock=self.lock)
        cmd.setParams(market, start)

        return cmd.call_api()

    @last_ack_time
    def get_minute_time_data(self, market, code):
        """
        获取分钟线
        :param market:
        :param code:
        :return:
        """
        cmd = GetMinuteTimeData(self.client, lock=self.lock)
        cmd.setParams(market, code)

        return cmd.call_api()

    @last_ack_time
    def get_history_minute_time_data(self, market, code, date):
        """
        获取分钟线历史数据
        :param market:
        :param code:
        :param date:
        :return:
        """
        cmd = GetHistoryMinuteTimeData(self.client, lock=self.lock)
        cmd.setParams(market, code, date)

        return cmd.call_api()

    @last_ack_time
    def get_transaction_data(self, market, code, start, count):
        """
        获取分笔数据
        :param market:
        :param code:
        :param start:
        :param count:
        :return:
        """
        cmd = GetTransactionData(self.client, lock=self.lock)
        cmd.setParams(market, code, start, count)

        return cmd.call_api()

    @last_ack_time
    def get_history_transaction_data(self, market, code, start, count, date):
        """
        分笔历史数据
        :param market:
        :param code:
        :param start:
        :param count:
        :param date:
        :return:
        """
        cmd = GetHistoryTransactionData(self.client, lock=self.lock)
        cmd.setParams(market, code, start, count, date)

        return cmd.call_api()

    @last_ack_time
    def get_company_info_category(self, market, code):
        """
        公司信息类别
        :param market:
        :param code:
        :return:
        """
        cmd = GetCompanyInfoCategory(self.client, lock=self.lock)
        cmd.setParams(market, code)

        return cmd.call_api()

    @last_ack_time
    def get_company_info_content(self, market, code, filename, start, length):
        """
        公司信息内容
        :param market:
        :param code:
        :param filename:
        :param start:
        :param length:
        :return:
        """
        cmd = GetCompanyInfoContent(self.client, lock=self.lock)
        cmd.setParams(market, code, filename, start, length)

        return cmd.call_api()

    @last_ack_time
    def get_xdxr_info(self, market, code):
        """
        除息除权信息
        :param market:
        :param code:
        :return:
        """
        cmd = GetXdXrInfo(self.client, lock=self.lock)
        cmd.setParams(market, code)

        return cmd.call_api()

    @last_ack_time
    def get_finance_info(self, market, code):
        """
        财务信息
        :param market:
        :param code:
        :return:
        """
        cmd = GetFinanceInfo(self.client, lock=self.lock)
        cmd.setParams(market, code)

        return cmd.call_api()

    @last_ack_time
    def get_block_info_meta(self, block_file):
        """
        板块元信息
        :param block_file:
        :return:
        """
        cmd = GetBlockInfoMeta(self.client, lock=self.lock)
        cmd.setParams(block_file)

        return cmd.call_api()

    @last_ack_time
    def get_block_info(self, block_file, start, size):
        """
        板块内容
        :param block_file:
        :param start:
        :param size:
        :return:
        """
        cmd = GetBlockInfo(self.client, lock=self.lock)
        cmd.setParams(block_file, start, size)

        return cmd.call_api()

    def get_and_parse_block_info(self, block_file):
        """
        下载并解析板块
        :param block_file:
        :return:
        """
        return get_and_parse_block_info(self, block_file)

    @last_ack_time
    def get_report_file(self, filename, offset):
        """
        下载财务报表文件
        :param filename:
        :param offset:
        :return:
        """
        cmd = GetReportFile(self.client, lock=self.lock)
        cmd.setParams(filename, offset)

        return cmd.call_api()

    def get_report_file_by_size(self, filename, filesize=0, reporthook=None):
        """
        Download file from proxy server
        :param reporthook:
        :param filename the filename to download
        :param filesize the file_size to download , if you do not know the actually file_size, leave this value 0
        """
        file_content = bytearray(filesize)
        get_zero_length_package_times = 0
        current_downloaded_size = 0

        while current_downloaded_size < filesize or filesize == 0:
            response = self.get_report_file(filename, current_downloaded_size)

            if response["chunksize"] > 0:
                current_downloaded_size = current_downloaded_size + response["chunksize"]
                file_content.extend(response["chunkdata"])

                reporthook and reporthook(current_downloaded_size, filesize)
            else:
                get_zero_length_package_times = get_zero_length_package_times + 1

                if filesize == 0 or get_zero_length_package_times > 2:
                    break

        return file_content

    def do_heartbeat(self):
        self.get_security_count(secrets.randbelow(1))

    def get_k_data(self, code, start_date, end_date):
        """
        巨 ugly 代码
        :param code:
        :param start_date:
        :param end_date:
        :return:
        """

        # 具体详情参见 https://github.com/rainx/pytdx/issues/5
        # 具体详情参见 https://github.com/rainx/pytdx/issues/21
        def __select_market_code(symbol):
            symbol = str(symbol)

            if symbol[0] in ["5", "6", "9"] or symbol[:3] in ["009", "126", "110", "201", "202", "203", "204"]:
                return 1

            return 0

        # 新版一劳永逸偷懒写法zzz
        # market_code = 1 if str(code)[0] == '6' else 0

        # https://github.com/rainx/pytdx/issues/33
        # 0 - 深圳， 1 - 上海

        result = pd.concat([self.to_df(self.get_security_bars(9, __select_market_code(code), code, (9 - i) * 800, 800)) for i in range(10)], axis=0)
        result = (
            result.assign(date=result["datetime"].apply(lambda x: str(x)[0:10]))
            .assign(code=str(code))
            .set_index("date", drop=False, inplace=False)
            .drop(["year", "month", "day", "hour", "minute", "datetime"], axis=1)[start_date:end_date]
        )

        result = result.assign(date=result["date"].apply(lambda x: str(x)[0:10]))

        return result



# cython: language_level=3

import struct
from pathlib import Path

from tdxpy.constants import SECURITY_COEFFICIENT
from tdxpy.logger import logger


def get_price(data, pos):
    """
    分析了一下，貌似是类似utf-8的编码方式保存有符号数字

    :param data:
    :param pos:
    :return:
    """

    pos_byte = 6

    bdata = index_bytes(data, pos)
    int_data = bdata & 0x3F

    if bdata & 0x40:
        sign = True
    else:
        sign = False

    if bdata & 0x80:
        while True:
            pos += 1

            bdata = index_bytes(data, pos)

            int_data += (bdata & 0x7F) << pos_byte
            pos_byte += 7

            if bdata & 0x80:
                pass
            else:
                break

    pos += 1

    if sign:
        int_data = -int_data

    return int_data, pos


def get_volume(vol):
    """
    获取交易量
    :param vol:
    :return:
    """
    logpoint = vol >> (8 * 3)

    hleax = (vol >> (8 * 2)) & 0xFF  # [2]
    lheax = (vol >> 8) & 0xFF  # [1]
    lleax = vol & 0xFF  # [0]

    dw_ecx = logpoint * 2 - 0x7F
    dw_edx = logpoint * 2 - 0x86

    dw_esi = logpoint * 2 - 0x8E
    dw_eax = logpoint * 2 - 0x96

    if dw_ecx < 0:
        tmp_eax = -dw_ecx
    else:
        tmp_eax = dw_ecx

    dbl_xmm6 = pow(2.0, tmp_eax)

    if dw_ecx < 0:
        dbl_xmm6 = 1.0 / dbl_xmm6

    if hleax > 0x80:
        dwtmpeax = dw_edx + 1
        tmpdbl_xmm3 = pow(2.0, dwtmpeax)

        dbl_xmm0 = pow(2.0, dw_edx) * 128.0
        dbl_xmm0 += (hleax & 0x7F) * tmpdbl_xmm3
        dbl_xmm4 = dbl_xmm0

    else:
        if dw_edx >= 0:
            dbl_xmm0 = pow(2.0, dw_edx) * hleax
        else:
            dbl_xmm0 = (1 / pow(2.0, dw_edx)) * hleax

        dbl_xmm4 = dbl_xmm0

    dbl_xmm3 = pow(2.0, dw_esi) * lheax
    dbl_xmm1 = pow(2.0, dw_eax) * lleax

    if hleax & 0x80:
        dbl_xmm3 *= 2.0
        dbl_xmm1 *= 2.0

    dbl_ret = dbl_xmm6 + dbl_xmm4 + dbl_xmm3 + dbl_xmm1

    return dbl_ret


def get_datetime(category, buffer, pos):
    """
    获取日期时间
    :param category:
    :param buffer:
    :param pos:
    :return:
    """
    minute = 0
    hour = 15

    if category < 4 or category == 7 or category == 8:
        zip_day, minutes = struct.unpack("<HH", buffer[pos: pos + 4])

        month = int((zip_day % 2048) / 100)
        year = (zip_day >> 11) + 2004
        day = (zip_day % 2048) % 100

        minute = minutes % 60
        hour = int(minutes / 60)
    else:
        zip_day, = struct.unpack("<I", buffer[pos: pos + 4])

        month = int((zip_day % 10000) / 100)
        year = int(zip_day / 10000)
        day = zip_day % 100

    pos += 4

    return year, month, day, hour, minute, pos


def get_time(buffer, pos):
    """
    获取时间
    :param buffer:
    :param pos:
    :return:
    """
    minutes, = struct.unpack("<H", buffer[pos: pos + 2])

    hour = int(minutes / 60)
    minute = minutes % 60

    pos += 2

    return hour, minute, pos


def index_bytes(data, pos):
    """
    索引比特
    :param data:
    :param pos:
    :return:
    """
    return data[pos]


def get_security_coefficient(market, name):
    return SECURITY_COEFFICIENT[get_security_type(market, name)]


def get_security_type(market, code):
    """
    获取股票类型, A股, B股, 指数等

    :param market: 市场
    :param code: 代码
    :return:
    """
    code = Path(code).stem
    code_head = code[:2]

    if market in ['SZ', 0]:
        if code_head in ["00", "30"]:
            return "SZ_A_STOCK"

        if code_head in ["20"]:
            return "SZ_B_STOCK"

        if code_head in ["39"]:
            return "SZ_INDEX"

        if code_head in ["15", "16"]:
            return "SZ_FUND"

        if code_head in ["10", "11", "12", "13", "14"]:
            return "SZ_BOND"

    if market in ['SH', 1]:
        if code_head in ["60", "68"]:  # 688XXX科创板
            return "SH_A_STOCK"

        if code_head in ["90"]:
            return "SH_B_STOCK"

        if code_head in ["00", "88", "99"]:
            return "SH_INDEX"

        if code_head in ["50", "51"]:
            return "SH_FUND"

        if code_head in ["01", "10", "11", "12", "13", "14", "20"]:
            return "SH_BOND"

    logger.debug("Unknown security exchange !")
    raise NotImplementedError


def dump(buf):
    from pprint import pprint

    try:
        from hexdump import hexdump
        pprint(hexdump(buf))
    except ImportError:
        pprint(buf)



# cython: language_level=3

import struct
from collections import OrderedDict

from tdxpy.logger import logger
from tdxpy.parser.base import BaseParser


class GetCompanyInfoCategory(BaseParser):
    def setParams(self, market, code):
        """
        设置参数
        :param market: 市场
        :param code: 代码
        """
        if type(code) is str:
            code = code.encode("utf-8")

        pkg = bytearray.fromhex("0c 0f 10 9b 00 01 0e 00 0e 00 cf 02")
        pkg.extend(struct.pack("<H6sI", market, code, 0))

        self.send_pkg = pkg

    def parseResponse(self, body_buf):
        """
        解析结果
        :param body_buf:
        :return:
        """
        pos = 0

        num, = struct.unpack("<H", body_buf[:2])

        pos += 2

        category = []

        def get_str(b):
            """
            :param b:
            :return:
            """
            p = b.find(b"\x00")

            if p != -1:
                b = b[0:p]

            try:
                n = b.decode("gbk")
            except Exception as e:
                logger.exception(e)
                n = "unknown_str"

            return n

        for _ in range(num):
            name, filename, start, length = struct.unpack("<64s80sII", body_buf[pos: pos + 152])
            pos += 152

            entry = OrderedDict([("name", get_str(name)), ("filename", get_str(filename)), ("start", start), ("length", length)])
            category.append(entry)

        return category



# cython: language_level=3

import struct
from collections import OrderedDict

from tdxpy.parser.base import BaseParser


class GetHistoryInstrumentBarsRange(BaseParser):
    def __init__(self, *args, **kvargs):
        super().__init__(self, *args, **kvargs)
        self.seq_id = 1

    def setParams(self, market, code, date, date2):
        """

        :param market:
        :param code:
        :param date:
        :param date2:
        """
        pkg = bytearray.fromhex("01")
        pkg.extend(struct.pack("<B", self.seq_id))
        pkg.extend(bytearray.fromhex("38 92 00 01 16 00 16 00 0D 24"))

        pkg.extend(struct.pack("<B9s", market, code.encode("utf-8")))
        pkg.extend(bytearray.fromhex("07 00"))
        pkg.extend(struct.pack("<LL", date, date2))

        self.send_pkg = pkg
        self.seq_id += 1

    @staticmethod
    def _parse_date(num):
        """

        :param num:
        :return:
        """
        month = (num % 2048) // 100
        year = num // 2048 + 2004
        day = (num % 2048) % 100

        return year, month, day

    @staticmethod
    def _parse_time(num):
        """

        :param num:
        :return:
        """
        return (num // 60), (num % 60)

    def parseResponse(self, body_buf):
        """

        :param body_buf:
        :return:
        """
        klines = []  # noqa
        pos = 12

        # 算了，前面不解析了，没太大用
        # (market, code) = struct.unpack("<B9s", body_buf[0: 10]

        ret_count, = struct.unpack("H", body_buf[pos: pos + 2])
        pos = pos + 2

        for _ in range(ret_count):
            (d1, d2, open_price, high, low, close, position, trade, settlementprice) = struct.unpack("<HHffffIIf", body_buf[pos: pos + 32])  # noqa
            pos += 32

            year, month, day = self._parse_date(d1)
            hour, minute = self._parse_time(d2)
            kline = OrderedDict(
                [
                    ("datetime", "%d-%02d-%02d %02d:%02d" % (year, month, day, hour, minute)),
                    ("year", year),
                    ("month", month),
                    ("day", day),
                    ("hour", hour),
                    ("minute", minute),
                    ("open", open_price),
                    ("high", high),
                    ("low", low),
                    ("close", close),
                    ("position", position),
                    ("trade", trade),
                    ("settlementprice", settlementprice),  # noqa
                ]
            )

            klines.append(kline)

        return klines



# cython: language_level=3

import secrets
import shutil
import tempfile
from pathlib import Path
from struct import calcsize
from struct import unpack

import pandas as pd

from .base_crawler import BaseCrawler
from ..exceptions import ValidationException

VALUE = "<6s1c1L"


class HistoryFinancialListCrawler(BaseCrawler):
    """
    获取历史财务数据的接口，参考上面issue里面 @datochan 的方案和代码
        https://github.com/rainx/tdxpy/issues/133
    """

    mode = "content"

    def get_url(self, *args, **kwargs):
        """

        :param args:
        :param kwargs:
        :return:
        """
        return "https://gitee.com/yutiansut/QADATA/raw/master/financial/content.txt"

    def get_content(self, reporthook=None, path_to_download=None, proxies=None, chunksize=1024 * 50, *args, **kwargs):
        """

        :param reporthook:
        :param path_to_download:
        :param proxies:
        :param chunksize:
        :param args:
        :param kwargs:
        :return:
        """
        from tdxpy.hq import TdxHq_API

        api = TdxHq_API()
        api.need_setup = False

        # calc.tdx.com.cn, calc2.tdx.com.cn
        with api.connect(ip="120.76.152.87"):
            content = api.get_report_file_by_size("tdxfin/gpcw.txt")
            download_file = not path_to_download and tempfile.NamedTemporaryFile(delete=True) or open(path_to_download, "wb")
            download_file.write(content)
            download_file.seek(0)

            return download_file

    def parse(self, download_file, *args, **kwargs):
        """

        :param download_file:
        :param args:
        :param kwargs:
        :return:
        """
        content = download_file.read()
        content = content.decode("utf-8")

        def list_to_dict(li):
            return {"filename": li[0], "hash": li[1], "filesize": int(li[2])}

        result = [list_to_dict(x) for x in [line.strip().split(",") for line in content.strip().split("\n")]]

        return result


class HistoryFinancialCrawler(BaseCrawler):
    mode = "content"

    def get_url(self, *args, **kwargs):
        """

        :param args:
        :param kwargs:
        :return:
        """
        if "filename" not in kwargs:
            raise ValidationException("Param filename is not set")

        filename = kwargs["filename"]

        return f"http://data.yutiansut.com/{filename}"  # noqa

    def get_content(self, reporthook=None, path_to_download=None, proxies=None, chunksize=1024 * 50, *args, **kwargs):
        """

        :param reporthook:
        :param path_to_download:
        :param proxies:
        :param chunksize:
        :param args:
        :param kwargs:
        :return:
        """
        from tdxpy.hq import TdxHq_API

        if "filename" not in kwargs:
            raise ValidationException("Param filename is not set")

        filename = kwargs["filename"]
        file_size = kwargs.get("filesize", 0)

        api = TdxHq_API()
        api.need_setup = False

        with api.connect(ip="120.76.152.87"):
            # calc.tdx.com.cn, calc2.tdx.com.cn
            content = api.get_report_file_by_size(f"tdxfin/{filename}", filesize=file_size, reporthook=reporthook)
            download_file = path_to_download and open(path_to_download, "wb") or tempfile.NamedTemporaryFile(delete=True)
            download_file.write(content)
            download_file.seek(0)

            return download_file

    def parse(self, download_file, *args, **kwargs):
        """

        :param download_file:
        :param args:
        :param kwargs:
        :return:
        """
        tmpdir = None
        header_pack_format = "<1hI1H3L"

        if download_file.name.endswith(".zip"):
            tmpdir_root = tempfile.gettempdir()
            subdir_name = f"tdxpy_{secrets.randbelow(1000000)}"

            tmpdir = Path(tmpdir_root, subdir_name)
            shutil.rmtree(tmpdir, ignore_errors=True)

            tmpdir.mkdir(parents=True, exist_ok=True)
            shutil.unpack_archive(download_file.name, extract_dir=tmpdir)

            # only one file endswith .dat should be in zip archives
            dat_file = None

            for _file in tmpdir.iterdir():
                if str(_file).endswith(".dat"):
                    dat_file = open(str(_file), "rb")

            if not dat_file:
                raise ValidationException("no dat file found in zip archive")
        else:
            dat_file = download_file

        header_size = calcsize(header_pack_format)
        stock_item_size = calcsize(VALUE)

        data_header = dat_file.read(header_size)
        stock_header = unpack(header_pack_format, data_header)

        max_count = stock_header[2]

        report_date = stock_header[1]
        report_size = stock_header[4]

        report_fields_count = int(report_size / 4)
        report_pack_format = "<{}f".format(report_fields_count)

        results = []

        for stock_idx in range(0, max_count):
            dat_file.seek(header_size + stock_idx * calcsize(VALUE))
            si = dat_file.read(stock_item_size)

            stock_item = unpack("<6s1c1L", si)
            code = stock_item[0].decode("utf-8")

            foa = stock_item[2]
            dat_file.seek(foa)

            info_data = dat_file.read(calcsize(report_pack_format))
            cw_info = unpack(report_pack_format, info_data)

            one_record = (code, report_date) + cw_info
            results.append(one_record)

        if download_file.name.endswith(".zip"):
            dat_file.close()
            shutil.rmtree(tmpdir, ignore_errors=True)

        return results

    @staticmethod
    def to_df(data):

        if not data:
            return None

        col = ["code", "report_date"]
        col += [f"col{str(i).zfill(3)}" for i in range(1, len(data[0]) - 1)]

        df = pd.DataFrame(data=data, columns=col)
        df.set_index("code", inplace=True)

        return df



# cython: language_level=3

import struct
from collections import OrderedDict

from tdxpy.parser.base import BaseParser


class GetInstrumentQuote(BaseParser):
    def setParams(self, market, code):
        """

        :param market:
        :param code:
        """
        pkg = bytearray.fromhex("01 01 08 02 02 01 0c 00 0c 00 fa 23")
        pkg.extend(struct.pack("<B9s", market, code.encode("utf-8")))

        self.send_pkg = pkg

    def parseResponse(self, body_buf):
        """

        :param body_buf:
        :return:
        """
        if len(body_buf) < 20:
            return []

        pos = 0

        market, code = struct.unpack("<B9s", body_buf[pos: pos + 10])

        pos += 10
        pos += 4

        # 持仓 ((13340,), 66),
        (
            pre_close,
            open_price,
            high,
            low,
            price,
            kaicang,
            _,
            zongliang,
            xianliang,
            _,
            neipan,
            waipai,
            _,
            chicang,
            b1,
            b2,
            b3,
            b4,
            b5,
            bv1,
            bv2,
            bv3,
            bv4,
            bv5,
            a1,
            a2,
            a3,
            a4,
            a5,
            av1,
            av2,
            av3,
            av4,
            av5,
        ) = struct.unpack("<fffffIIIIIIIIIfffffIIIIIfffffIIIII", body_buf[pos: pos + 136])

        return [
            OrderedDict(
                [
                    ("market", market),
                    ("code", code.decode("utf-8").rstrip("\x00")),
                    ("pre_close", pre_close),
                    ("open", open_price),
                    ("high", high),
                    ("low", low),
                    ("price", price),
                    ("kaicang", kaicang),
                    ("zongliang", zongliang),
                    ("xianliang", xianliang),
                    ("neipan", neipan),
                    ("waipan", waipai),
                    ("chicang", chicang),
                    ("bid1", b1),
                    ("bid2", b2),
                    ("bid3", b3),
                    ("bid4", b4),
                    ("bid5", b5),
                    ("bid_vol1", bv1),
                    ("bid_vol2", bv2),
                    ("bid_vol3", bv3),
                    ("bid_vol4", bv4),
                    ("bid_vol5", bv5),
                    ("ask1", a1),
                    ("ask2", a2),
                    ("ask3", a3),
                    ("ask4", a4),
                    ("ask5", a5),
                    ("ask_vol1", av1),
                    ("ask_vol2", av2),
                    ("ask_vol3", av3),
                    ("ask_vol4", av4),
                    ("ask_vol5", av5),
                ]
            )
        ]



# cython: language_level=3

import struct
from collections import OrderedDict

from tdxpy.helper import get_datetime
from tdxpy.helper import get_price
from tdxpy.helper import get_volume
from tdxpy.parser.base import BaseParser


class GetIndexBarsCmd(BaseParser):
    def setParams(self, category, market, code, start, count):
        """

        :param category:
        :param market:
        :param code:
        :param start:
        :param count:
        """
        if type(code) is str:
            code = code.encode("utf-8")

        self.category = category

        val = (0x10C, 0x01016408, 0x1C, 0x1C, 0x052D, market, code, category, 1, start, count, 0, 0, 0)  # I + I +  H total 10 zero
        pkg = struct.pack("<HIHHHH6sHHHHIIH", *val)  # noqa

        self.send_pkg = pkg

    def parseResponse(self, body_buf):
        """

        :param body_buf:
        :return:
        """
        ret_count, = struct.unpack("<H", body_buf[0:2])

        pos = 2

        klines = []  # noqa
        pre_diff_base = 0

        for _ in range(ret_count):
            year, month, day, hour, minute, pos = get_datetime(self.category, body_buf, pos)  # noqa

            price_open_diff, pos = get_price(body_buf, pos)
            price_close_diff, pos = get_price(body_buf, pos)

            price_high_diff, pos = get_price(body_buf, pos)
            price_low_diff, pos = get_price(body_buf, pos)

            vol_raw, = struct.unpack("<I", body_buf[pos: pos + 4])
            vol = get_volume(vol_raw)

            pos += 4

            db_vol_raw, = struct.unpack("<I", body_buf[pos: pos + 4])
            db_vol = get_volume(db_vol_raw)

            pos += 4

            (up_count, down_count) = struct.unpack("<HH", body_buf[pos: pos + 4])

            pos += 4  # noqa

            open_ = self._cal_price1000(price_open_diff, pre_diff_base)
            price_open_diff = price_open_diff + pre_diff_base

            close = self._cal_price1000(price_open_diff, price_close_diff)
            high = self._cal_price1000(price_open_diff, price_high_diff)
            low = self._cal_price1000(price_open_diff, price_low_diff)

            pre_diff_base = price_open_diff + price_close_diff

            # 为了避免python处理浮点数的时候，浮点数运算不精确问题，这里引入了多余的代码
            kline = OrderedDict(
                [
                    ("open", open_),
                    ("close", close),
                    ("high", high),
                    ("low", low),
                    ("vol", vol),
                    ("amount", db_vol),
                    ("year", year),
                    ("month", month),
                    ("day", day),
                    ("hour", hour),
                    ("minute", minute),
                    ("datetime", f"{year:d}-{month:02d}-{day:02d} {hour:02d}:{minute:02d}"),
                    ("up_count", up_count),
                    ("down_count", down_count),
                ]
            )

            klines.append(kline)

        return klines



# cython: language_level=3

import struct
from collections import OrderedDict

from tdxpy.parser.base import BaseParser


class GetInstrumentInfo(BaseParser):

    def setParams(self, start, count=100):
        """

        :param start:
        :param count:
        """
        pkg = bytearray.fromhex("01 04 48 67 00 01 08 00 08 00 f5 23")
        pkg.extend(struct.pack("<IH", start, count))

        self.send_pkg = pkg

    def parseResponse(self, body_buf):
        """

        :param body_buf:
        :return:
        """
        pos = 0
        start, count = struct.unpack("<IH", body_buf[:6])

        pos += 6
        result = []

        for _ in range(count):
            category, market, _, code_raw, name_raw, desc_raw = struct.unpack("<BB3s9s17s9s", body_buf[pos: pos + 40])

            code = code_raw.decode("gbk", "ignore").rstrip("\x00")
            name = name_raw.decode("gbk", "ignore").rstrip("\x00")
            desc = desc_raw.decode("gbk", "ignore").rstrip("\x00")

            line = OrderedDict(
                [
                    ("category", category),
                    ("market", market),
                    ("code", code),
                    ("name", name),
                    ("desc", desc),
                ]
            )

            result.append(line)

            pos += 64

        return result



# cython: language_level=3

import struct

from tdxpy.parser.base import BaseParser


class GetInstrumentCount(BaseParser):
    def setup(self):
        self.send_pkg = bytearray.fromhex("01 03 48 66 00 01 02 00 02 00 f0 23")

    def parseResponse(self, body_buf):
        num, = struct.unpack("<I", body_buf[19: 19 + 4])

        return num



# cython: language_level=3

import datetime
import struct
from collections import OrderedDict

from tdxpy.parser.base import BaseParser


class GetTransactionData(BaseParser):
    def setParams(self, market, code, start, count):
        """

        :param market:
        :param code:
        :param start:
        :param count:
        """
        if type(code) is str:
            code = code.encode("utf-8")

        pkg = bytearray.fromhex("01 01 08 00 03 01 12 00 12 00 fc 23")
        pkg.extend(struct.pack("<B9siH", market, code, start, count))

        self.send_pkg = pkg

    def parseResponse(self, body_buf):
        """

        :param body_buf:
        :return:
        """
        pos = 0
        market, code, _, num = struct.unpack("<B9s4sH", body_buf[pos: pos + 16])

        pos += 16
        result = []
        nature_name = ""

        for _ in range(num):

            raw_time, price, volume, zengcang, direction = struct.unpack("<HIIiH", body_buf[pos: pos + 16])

            pos += 16
            hour = raw_time // 60

            minute = raw_time % 60
            second = direction % 10000
            nature = direction  # 保持老接口的兼容性

            if second > 59:
                second = 0

            date = datetime.datetime.combine(datetime.date.today(), datetime.time(hour, minute, second))

            value = direction // 10000

            if value == 0:
                direction = 1
                if zengcang > 0:
                    if volume > zengcang:
                        nature_name = "多开"
                    elif volume == zengcang:
                        nature_name = "双开"
                elif zengcang == 0:
                    nature_name = "多换"
                else:
                    if volume == -zengcang:
                        nature_name = "双平"
                    else:
                        nature_name = "空平"
            elif value == 1:
                direction = -1
                if zengcang > 0:
                    if volume > zengcang:
                        nature_name = "空开"
                    elif volume == zengcang:
                        nature_name = "双开"
                elif zengcang == 0:
                    nature_name = "空换"
                else:
                    if volume == -zengcang:
                        nature_name = "双平"
                    else:
                        nature_name = "多平"
            else:
                direction = 0
                if zengcang > 0:
                    if volume > zengcang:
                        nature_name = "开仓"
                    elif volume == zengcang:
                        nature_name = "双开"
                elif zengcang < 0:
                    if volume > -zengcang:
                        nature_name = "平仓"
                    elif volume == -zengcang:
                        nature_name = "双平"
                else:
                    nature_name = "换手"

            if market in [31, 48]:
                if nature == 0:
                    direction = 1
                    nature_name = "B"
                elif nature == 256:
                    direction = -1
                    nature_name = "S"
                else:  # 512
                    direction = 0
                    nature_name = ""

            result.append(
                OrderedDict(
                    [
                        ("date", date),
                        ("hour", hour),
                        ("minute", minute),
                        ("second", second),
                        ("price", price),
                        ("volume", volume),
                        ("zengcang", zengcang),
                        ("nature", nature),
                        ("nature_mark", nature // 10000),
                        ("nature_value", nature % 10000),
                        ("nature_name", nature_name),
                        ("direction", direction),
                    ]
                )
            )

        return result



# cython: language_level=3

import struct
from collections import OrderedDict

from tdxpy.parser.base import BaseParser


class GetHistoryMinuteTimeData(BaseParser):
    def setParams(self, market, code, date):
        pkg = bytearray.fromhex("01 01 30 00 01 01 10 00 10 00 0c 24")
        pkg.extend(struct.pack("<IB9s", date, market, code.encode("utf-8")))

        self.send_pkg = pkg

    def parseResponse(self, body_buf):
        pos = 0

        market, code, _, num = struct.unpack("<B9s8sH", body_buf[pos: pos + 20])

        pos += 20

        result = []

        for _ in range(num):
            raw_time, price, avg_price, volume, amount = struct.unpack("<HffII", body_buf[pos: pos + 18])

            pos += 18

            hour = raw_time // 60
            minute = raw_time % 60

            result.append(
                OrderedDict(
                    [
                        ("hour", hour),
                        ("minute", minute),
                        ("price", price),
                        ("avg_price", avg_price),
                        ("volume", volume),
                        ("open_interest", amount),
                    ]
                )
            )

        return result



# cython: language_level=3

import struct
from collections import OrderedDict

from tdxpy.helper import get_price
from tdxpy.parser.base import BaseParser


class GetHistoryMinuteTimeData(BaseParser):
    def setParams(self, market, code, date):
        """
        设置参数
        :param market: 0/1 股票市场
        :param code: '000001' 股票代码
        :param date: 20161201  类似这样的整型
        :return:
        """

        if (type(date) is str) or (type(date) is bytes):
            date = int(date)

        if type(code) is str:
            code = code.encode("utf-8")

        pkg = bytearray.fromhex("0c 01 30 00 01 01 0d 00 0d 00 b4 0f")
        pkg.extend(struct.pack("<IB6s", date, market, code))

        self.send_pkg = pkg

    def parseResponse(self, body_buf):
        """

        :param body_buf:
        :return:
        """
        pos = 0

        num, = struct.unpack("<H", body_buf[:2])
        last_price = 0

        # 跳过了4个字节，实在不知道是什么意思
        pos += 6
        prices = []

        for _ in range(num):
            price_raw, pos = get_price(body_buf, pos)
            reversed1, pos = get_price(body_buf, pos)

            vol, pos = get_price(body_buf, pos)
            last_price = last_price + price_raw

            price = OrderedDict([("price", float(last_price) / 100), ("vol", vol)])
            prices.append(price)

        return prices



#!/usr/bin/env python
from setuptools import find_packages
from setuptools import setup

cython_deps = ["cython", ]
ext_modules = {}

# try:
#     from Cython.Build import cythonize
#
#     ext_modules = {
#         "ext_modules": cythonize(
#             [
#                 "tdxpy/parser/std/get_security_quotes.py",
#                 "tdxpy/base_socket_client.py",
#                 "tdxpy/parser/base.py",
#                 "tdxpy/helper.py",
#                 "tdxpy/hq.py",
#             ]
#         )
#     }
#
# except ImportError:
#     pass

try:
    import pypandoc
    long_description = pypandoc.convert_file('README.md', 'rst')
except (IOError, ImportError):
    print(30 * "*")
    print("Notice, NEED TO INSTALL *pypandoc* TO get full description of package")
    print(30 * "*")
    long_description = ''
    exit()

setup(
    name="tdxpy",
    version="0.1.6",
    description="TdxPy - Python TDX 数据接口",
    long_description=long_description,
    maintainer="BoPo",
    maintainer_email="ibopo@126.com",
    url="https://github.com/mootdx/tdxpy",
    packages=find_packages(include=["tdxpy", "tdxpy.*"]),
    include_package_data=True,
    install_requires=[
        "cryptography>=38.0.3",
        "pandas>=1.5.1",
    ],
    # extras_require={
    #     # can be installed by pip install modin[dask]
    #     "cython": cython_deps,
    # },
    # **ext_modules,
)



# cython: language_level=3

import struct
from collections import OrderedDict

from tdxpy.logger import logger
from tdxpy.parser.base import BaseParser


class GetInstrumentQuoteList(BaseParser):
    def setParams(self, market, category, start, count):
        """

        :param market:
        :param category:
        :param start:
        :param count:
        """
        pkg = bytearray.fromhex("01 c1 06 0b 00 02 0b 00 0b 00 00 24")
        pkg.extend(struct.pack("<BHHHH", market, 0, start, count, 1))

        self.category = category
        self.send_pkg = pkg

    def parseResponse(self, body_buf):
        """

        :param body_buf:
        :return:
        """
        pos = 0

        data_list = []

        num, = struct.unpack("<H", body_buf[pos: pos + 2])

        if num == 0:
            return []

        pos += 2

        if self.category not in [2, 3]:
            return NotImplementedError("暂时不支持期货,港股之外的品类")

        for _ in range(num):
            # 每个块一共300bytes
            market, code = struct.unpack("<B9s", body_buf[pos: pos + 10])
            code = code.strip(b"\0").decode("gbk")  # to unicode
            pos += 10

            if self.category == 3:
                try:
                    pos = self.extract_futures(market, code, body_buf, data_list, pos)
                except Exception as e:
                    logger.exception(e)

                logger.debug(pos)

            elif self.category == 2:
                pos = self.extract_hongkong_stocks(market, code, body_buf, data_list, pos)

        return data_list

    @staticmethod
    def extract_hongkong_stocks(market, code, body_buf, datalist, pos):
        """

        :param market:
        :param code:
        :param body_buf:
        :param datalist:
        :param pos:
        :return:
        """
        data_pack_format = "<IfffffIfIIfIIIIfffffIIIIIfffffIIIII"
        (
            HuoYueDu,
            ZuoShou,
            JinKai,
            ZuiGao,
            ZuiDi,
            XianJia,
            _,  # 0
            MaiRuJia,  # ?
            ZongLiang,
            XianLiang,  # ?
            ZongJinE,
            _,  # ?
            _,  # ?
            Nei,  # 0
            Wai,  # 0 Nei/Wai = 内外比？
            MaiRuJia1,
            MaiRuJia2,
            MaiRuJia3,
            MaiRuJia4,
            MaiRuJia5,
            MaiRuLiang1,
            MaiRuLiang2,
            MaiRuLiang3,
            MaiRuLiang4,
            MaiRuLiang5,
            MaiChuJia1,
            MaiChuJia2,
            MaiChuJia3,
            MaiChuJia4,
            MaiChuJia5,
            MaiChuLiang1,
            MaiChuLiang2,
            MaiChuLiang3,
            MaiChuLiang4,
            MaiChuLiang5,
        ) = struct.unpack(data_pack_format, body_buf[pos: pos + 140])

        pos += 290

        one = OrderedDict(
            [
                ("market", market),
                ("code", code),
                ("HuoYueDu", HuoYueDu),
                ("ZuoShou", ZuoShou),
                ("JinKai", JinKai),
                ("ZuiGao", ZuiGao),
                ("ZuiDi", ZuiDi),
                ("XianJia", XianJia),
                ("MaiRuJia", MaiRuJia),
                ("ZongLiang", ZongLiang),
                ("XianLiang", XianLiang),
                ("ZongJinE", ZongJinE),
                ("Nei", Nei),
                ("Wai", Wai),  # 0 Nei/Wai = 内外比？
                ("MaiRuJia1", MaiRuJia1),
                ("MaiRuJia2", MaiRuJia2),
                ("MaiRuJia3", MaiRuJia3),
                ("MaiRuJia4", MaiRuJia4),
                ("MaiRuJia5", MaiRuJia5),
                ("MaiRuLiang1", MaiRuLiang1),
                ("MaiRuLiang2", MaiRuLiang2),
                ("MaiRuLiang3", MaiRuLiang3),
                ("MaiRuLiang4", MaiRuLiang4),
                ("MaiRuLiang5", MaiRuLiang5),
                ("MaiChuJia1", MaiChuJia1),
                ("MaiChuJia2", MaiChuJia2),
                ("MaiChuJia3", MaiChuJia3),
                ("MaiChuJia4", MaiChuJia4),
                ("MaiChuJia5", MaiChuJia5),
                ("MaiChuLiang1", MaiChuLiang1),
                ("MaiChuLiang2", MaiChuLiang2),
                ("MaiChuLiang3", MaiChuLiang3),
                ("MaiChuLiang4", MaiChuLiang4),
                ("MaiChuLiang5", MaiChuLiang5),
            ]
        )

        datalist.append(one)

        return pos

    @staticmethod
    def extract_futures(market, code, body_buf, datalist, pos):
        """

        :param market:
        :param code:
        :param body_buf:
        :param datalist:
        :param pos:
        :return:
        """

        data_pack_format = "<IfffffIIIIfIIfIfIIIIIIIIIfIIIIIIIII"

        (
            BiShu,
            ZuoJie,
            JinKai,
            ZuiGao,
            ZuiDi,
            MaiChu,
            KaiCang,
            _,
            ZongLiang,
            XianLiang,
            ZongJinE,
            NeiPan,
            WaiPan,
            _,
            ChiCangLiang,
            MaiRuJia,
            _,
            _,
            _,
            _,
            MaiRuLiang,
            _,
            _,
            _,
            _,
            MaiChuJia,
            _,
            _,
            _,
            _,
            MaiChuLiang,
            _,
            _,
            _,
            _,
        ) = struct.unpack(data_pack_format, body_buf[pos: pos + 140])

        pos += 290

        one = OrderedDict(
            [
                ("market", market),
                ("code", code),
                ("BiShu", BiShu),
                ("ZuoJie", ZuoJie),
                ("JinKai", JinKai),
                ("ZuiGao", ZuiGao),
                ("ZuiDi", ZuiDi),
                ("MaiChu", MaiChu),
                ("KaiCang", KaiCang),
                ("ZongLiang", ZongLiang),
                ("XianLiang", XianLiang),
                ("ZongJinE", ZongJinE),
                ("NeiPan", NeiPan),
                ("WaiPan", WaiPan),
                ("ChiCangLiang", ChiCangLiang),
                ("MaiRuJia", MaiRuJia),
                ("MaiRuLiang", MaiRuLiang),
                ("MaiChuJia", MaiChuJia),
                ("MaiChuLiang", MaiChuLiang),
            ]
        )

        datalist.append(one)

        return datalist



# cython: language_level=3
import abc
import datetime
import struct

import zlib

from tdxpy.logger import logger

try:
    import cython  # noqa

    if cython.compiled:
        def buffer(x):
            return x

except ImportError:
    pass


class SocketClientNotReady(Exception):
    pass


class SendPkgNotReady(Exception):
    pass


class SendRequestPkgFails(Exception):
    pass


class ResponseHeaderRecvFails(Exception):
    pass


class ResponseRecvFails(Exception):
    pass


RSP_HEADER_LEN = 0x10


class BaseParser(object):
    def __init__(self, client, lock=None):
        """

        :rtype: object
        """

        self.send_pkg = None
        self.data = None

        self.rsp_header_len = RSP_HEADER_LEN
        self.rsp_header = None
        self.rsp_body = None

        self.client = client
        self.lock = lock or None

        self.category = None

    def setParams(self, *args, **xargs):  # noqa
        """
        构建请求
        :return:
        """
        pass

    @abc.abstractmethod
    def parseResponse(self, body_buf):  # noqa
        """
        解析结果
        :param body_buf:
        """
        pass

    @staticmethod
    def _parse_date(num):
        """

        :param num:
        :return:
        """
        month = (num % 2048) // 100
        year = num // 2048 + 2004
        day = (num % 2048) % 100

        return year, month, day

    @staticmethod
    def _parse_time(num):
        """

        :param num:
        :return:
        """
        return (num // 60), (num % 60)

    @staticmethod
    def _cal_price1000(base_p, diff):
        return float(base_p + diff) / 1000

    def setup(self):
        pass

    def call_api(self):
        """

        :return:
        """
        if self.lock:
            with self.lock:
                logger.debug("sending thread lock api call")
                result = self._call_api()
        else:
            result = self._call_api()

        return result

    def _call_api(self):
        """

        :return:
        """
        self.setup()

        if not self.client:
            raise SocketClientNotReady("socket client not ready")

        if not self.send_pkg:
            raise SendPkgNotReady("send pkg not ready")

        # 发送数据包
        descended = self.client.send(self.send_pkg)

        logger.debug(f"descended, {descended}")

        self.client.send_pkg_num += 1
        self.client.send_pkg_bytes += descended
        self.client.last_api_send_bytes = descended

        if not self.client.first_pkg_send_time:
            self.client.first_pkg_send_time = datetime.datetime.now()

        if descended != len(self.send_pkg):
            logger.debug("send bytes error")
            raise SendRequestPkgFails("send fails")
        else:
            # 接收数据
            logger.debug("send rsp_header_len:" + str(self.rsp_header_len))
            head_buf = self.client.recv(self.rsp_header_len)

            # 长度符合
            if len(head_buf) == self.rsp_header_len:
                self.client.recv_pkg_num += 1
                self.client.recv_pkg_bytes += self.rsp_header_len

                # 解包
                _, _, _, zip_size, unzip_size = struct.unpack("<IIIHH", head_buf)

                logger.debug(f"zip size is: {zip_size}")
                logger.debug(f"unzip size is: {unzip_size}")

                body_buf = bytearray()
                last_api_recv_bytes = self.rsp_header_len

                while True:
                    # 接收数据包
                    buf = self.client.recv(zip_size)
                    len_buf = len(buf)

                    self.client.recv_pkg_num += 1
                    self.client.recv_pkg_bytes += len_buf

                    last_api_recv_bytes += len_buf
                    body_buf.extend(buf)

                    # 接收数据到结束
                    if not buf or len_buf == 0 or len(body_buf) == zip_size:
                        logger.debug("接收数据到结束")
                        break

                self.client.last_api_recv_bytes = last_api_recv_bytes

                if len(buf) == 0:
                    logger.debug("接收数据体失败服务器断开连接")
                    raise ResponseRecvFails("接收数据体失败服务器断开连接")

                if zip_size != unzip_size:
                    logger.debug("需要解压, 解压数据")
                    body_buf = zlib.decompress(body_buf)

                return self.parseResponse(body_buf)

            else:
                # 长度不符合抛出异常
                logger.debug("head_buf is not 0x10")
                raise ResponseHeaderRecvFails("head_buf is not 0x10 : " + str(head_buf))



# cython: language_level=3

from tdxpy.base_socket_client import BaseSocketClient
from tdxpy.base_socket_client import last_ack_time
from tdxpy.parser.ext.ex_get_history_instrument_bars_range import GetHistoryInstrumentBarsRange
from tdxpy.parser.ext.ex_get_history_minute_time_data import GetHistoryMinuteTimeData
from tdxpy.parser.ext.ex_get_history_transaction_data import GetHistoryTransactionData
from tdxpy.parser.ext.ex_get_instrument_bars import GetInstrumentBars
from tdxpy.parser.ext.ex_get_instrument_count import GetInstrumentCount
from tdxpy.parser.ext.ex_get_instrument_info import GetInstrumentInfo
from tdxpy.parser.ext.ex_get_instrument_quote import GetInstrumentQuote
from tdxpy.parser.ext.ex_get_instrument_quote_list import GetInstrumentQuoteList
from tdxpy.parser.ext.ex_get_markets import GetMarkets
from tdxpy.parser.ext.ex_get_minute_time_data import GetMinuteTimeData
from tdxpy.parser.ext.ex_get_transaction_data import GetTransactionData
from tdxpy.parser.setup_commands import ExSetupCmd1

"""
In [7]: 0x7e
Out[7]: 126

In [5]: len(body)
Out[5]: 8066

In [6]: len(body)/126
Out[6]: 64.01587301587301

In [7]: len(body)%126
Out[7]: 2

In [8]: (len(body)-2)/126
Out[8]: 64.0
"""


class TdxExHq_API(BaseSocketClient):  # noqa
    def setup(self):
        ExSetupCmd1(self.client).call_api()

    @last_ack_time
    def get_markets(self):
        """
        获得市场列表
        :return:
        """
        cmd = GetMarkets(self.client)
        return cmd.call_api()

    @last_ack_time
    def get_instrument_count(self):
        """
        获取证券数量
        :return:
        """
        cmd = GetInstrumentCount(self.client)
        return cmd.call_api()

    @last_ack_time
    def get_instrument_quote(self, market, code):
        """
        获取行情
        :param market:
        :param code:
        :return:
        """
        cmd = GetInstrumentQuote(self.client)
        cmd.setParams(market, code)
        return cmd.call_api()

    @last_ack_time
    def get_instrument_bars(self, category, market, code, start=0, count=700):
        """
        获取时间线
        :param category:
        :param market:
        :param code:
        :param start:
        :param count:
        :return:
        """
        cmd = GetInstrumentBars(self.client)
        cmd.setParams(category, market, code, start=start, count=count)
        return cmd.call_api()

    @last_ack_time
    def get_minute_time_data(self, market, code):
        """
        分钟线

        :param market:
        :param code:
        :return:
        """
        cmd = GetMinuteTimeData(self.client)
        cmd.setParams(market, code)
        return cmd.call_api()

    @last_ack_time
    def get_history_minute_time_data(self, market, code, date):
        """
        分钟线历史
        :param market:
        :param code:
        :param date:
        :return:
        """
        cmd = GetHistoryMinuteTimeData(self.client)
        cmd.setParams(market, code, date=date)
        return cmd.call_api()

    @last_ack_time
    def get_transaction_data(self, market, code, start=0, count=1800):
        """
        分笔数据
        :param market:
        :param code:
        :param start:
        :param count:
        :return:
        """
        cmd = GetTransactionData(self.client)
        cmd.setParams(market, code, start=start, count=count)
        return cmd.call_api()

    @last_ack_time
    def get_history_transaction_data(self, market, code, date, start=0, count=1800):
        """
        分笔历史数据
        :param market:
        :param code:
        :param date:
        :param start:
        :param count:
        :return:
        """
        cmd = GetHistoryTransactionData(self.client)
        cmd.setParams(market, code, date, start=start, count=count)
        return cmd.call_api()

    @last_ack_time
    def get_history_instrument_bars_range(self, market, code, start, end):
        """

        :param market:
        :param code:
        :param start:
        :param end:
        :return:
        """
        cmd = GetHistoryInstrumentBarsRange(self.client)
        cmd.setParams(market, code, start, end)
        return cmd.call_api()

    @last_ack_time
    def get_instrument_info(self, start, count=100):
        """

        :param start:
        :param count:
        :return:
        """
        cmd = GetInstrumentInfo(self.client)
        cmd.setParams(start, count)
        return cmd.call_api()

    @last_ack_time
    def get_instrument_quote_list(self, market, category, start=0, count=80):
        """

        :param market:
        :param category:
        :param start:
        :param count:
        :return:
        """
        cmd = GetInstrumentQuoteList(self.client)
        cmd.setParams(market, category, start, count)
        return cmd.call_api()

    def do_heartbeat(self):
        self.get_instrument_count()



# cython: language_level=3

import struct
from collections import OrderedDict

from tdxpy.parser.base import BaseParser


class GetFinanceInfo(BaseParser):
    def setParams(self, market, code):
        """

        :param market:
        :param code:
        """
        if type(code) is str:
            code = code.encode("utf-8")

        pkg = bytearray.fromhex("0c 1f 18 76 00 01 0b 00 0b 00 10 00 01 00")
        pkg.extend(struct.pack("<B6s", market, code))

        self.send_pkg = pkg

    def parseResponse(self, body_buf):
        """

        :param body_buf:
        :return:
        """
        pos = 0
        pos += 2  # skip num ,we only query 1 in this case

        market, code = struct.unpack("<B6s", body_buf[pos: pos + 7])

        pos += 7

        (
            liutongguben,
            province,
            industry,
            updated_date,
            ipo_date,
            zongguben,
            guojiagu,
            faqirenfarengu,
            farengu,
            bgu,
            hgu,
            zhigonggu,
            zongzichan,
            liudongzichan,
            gudingzichan,
            wuxingzichan,
            gudongrenshu,
            liudongfuzhai,
            changqifuzhai,
            zibengongjijin,
            jingzichan,
            zhuyingshouru,
            zhuyinglirun,
            yingshouzhangkuan,
            yingyelirun,
            touzishouyu,
            jingyingxianjinliu,
            zongxianjinliu,
            cunhuo,
            lirunzonghe,
            shuihoulirun,
            jinglirun,
            weifenlirun,
            baoliu1,
            baoliu2,
        ) = struct.unpack("<fHHIIffffffffffffffffffffffffffffff", body_buf[pos:])

        def _get_v(v):
            return v

        result = OrderedDict(
            [
                ("market", market),
                ("code", code.decode("utf-8")),
                ("liutongguben", _get_v(liutongguben) * 10000),
                ("province", province),
                ("industry", industry),
                ("updated_date", updated_date),
                ("ipo_date", ipo_date),
                ("zongguben", _get_v(zongguben) * 10000),
                ("guojiagu", _get_v(guojiagu) * 10000),
                ("faqirenfarengu", _get_v(faqirenfarengu) * 10000),
                ("farengu", _get_v(farengu) * 10000),
                ("bgu", _get_v(bgu) * 10000),
                ("hgu", _get_v(hgu) * 10000),
                ("zhigonggu", _get_v(zhigonggu) * 10000),
                ("zongzichan", _get_v(zongzichan) * 10000),
                ("liudongzichan", _get_v(liudongzichan) * 10000),
                ("gudingzichan", _get_v(gudingzichan) * 10000),
                ("wuxingzichan", _get_v(wuxingzichan) * 10000),
                ("gudongrenshu", _get_v(gudongrenshu)),
                ("liudongfuzhai", _get_v(liudongfuzhai) * 10000),
                ("changqifuzhai", _get_v(changqifuzhai) * 10000),
                ("zibengongjijin", _get_v(zibengongjijin) * 10000),
                ("jingzichan", _get_v(jingzichan) * 10000),
                ("zhuyingshouru", _get_v(zhuyingshouru) * 10000),
                ("zhuyinglirun", _get_v(zhuyinglirun) * 10000),
                ("yingshouzhangkuan", _get_v(yingshouzhangkuan) * 10000),
                ("yingyelirun", _get_v(yingyelirun) * 10000),
                ("touzishouyu", _get_v(touzishouyu) * 10000),
                ("jingyingxianjinliu", _get_v(jingyingxianjinliu) * 10000),
                ("zongxianjinliu", _get_v(zongxianjinliu) * 10000),
                ("cunhuo", _get_v(cunhuo) * 10000),
                ("lirunzonghe", _get_v(lirunzonghe) * 10000),
                ("shuihoulirun", _get_v(shuihoulirun) * 10000),
                ("jinglirun", _get_v(jinglirun) * 10000),
                ("weifenpeilirun", _get_v(weifenlirun) * 10000),
                ("meigujingzichan", _get_v(baoliu1)),
                ("baoliu2", _get_v(baoliu2)),
            ]
        )

        return result



# cython: language_level=3

import struct
from collections import OrderedDict

from tdxpy.helper import get_datetime
from tdxpy.parser.base import BaseParser


class GetInstrumentBars(BaseParser):

    def setParams(self, category, market, code, start, count):
        """

        :param category:
        :param market:
        :param code:
        :param start:
        :param count:
        """
        if type(code) is str:
            code = code.encode("utf-8")

        pkg = bytearray.fromhex("01 01 08 6a 01 01 16 00 16 00")
        pkg.extend(bytearray.fromhex("ff 23"))

        self.category = category
        pkg.extend(struct.pack("<B9sHHIH", market, code, category, 1, start, count))

        # 这个1还不确定是什么作用，疑似和是否复权有关
        self.send_pkg = pkg

    def parseResponse(self, body_buf):
        """

        :param body_buf:
        :return:
        """
        pos = 0
        pos += 18

        ret_count, = struct.unpack("<H", body_buf[pos: pos + 2])
        pos += 2

        klines = []

        for _ in range(ret_count):
            year, month, day, hour, minute, pos = get_datetime(self.category, body_buf, pos)
            open_price, high, low, close, position, trade, price = struct.unpack("<ffffIIf", body_buf[pos: pos + 28])

            amount, = struct.unpack("f", body_buf[pos + 16: pos + 16 + 4])
            pos += 28

            kline = OrderedDict(
                [
                    ("open", open_price),
                    ("high", high),
                    ("low", low),
                    ("close", close),
                    ("position", position),
                    ("trade", trade),
                    ("price", price),
                    ("year", year),
                    ("month", month),
                    ("day", day),
                    ("hour", hour),
                    ("minute", minute),
                    ("datetime", "%d-%02d-%02d %02d:%02d" % (year, month, day, hour, minute)),
                    ("amount", amount),
                ]
            )

            klines.append(kline)

        return klines



# cython: language_level=3

import struct
from collections import OrderedDict

from tdxpy.helper import get_price
from tdxpy.helper import get_security_coefficient
from tdxpy.helper import get_volume
from tdxpy.parser.base import BaseParser


class GetSecurityQuotesCmd(BaseParser):
    def setParams(self, all_stock):
        """
        :param all_stock: 一个包含 (market, code) 元组的列表， 如 [ (0, '000001'), (1, '600001') ]
        :return:
        """
        stock_len = len(all_stock)

        if stock_len <= 0:
            return False

        pkgdatalen = stock_len * 7 + 12

        val = (0x10C, 0x02006320, pkgdatalen, pkgdatalen, 0x5053E, 0, 0, stock_len)
        pkg = bytearray(struct.pack("<HIHHIIHH", *val))

        for stock in all_stock:
            market, code = stock

            if type(code) is str:
                code = code.encode("utf-8")

            one_stock_pkg = struct.pack("<B6s", market, code)
            pkg.extend(one_stock_pkg)

        self.send_pkg = pkg

    def parseResponse(self, body_buf):
        """

        :param body_buf:
        :return:
        """
        pos = 0
        pos += 2  # skip b1 cb

        num_stock, = struct.unpack("<H", body_buf[pos: pos + 2])

        pos += 2
        stocks = []

        for _ in range(num_stock):
            # print(body_buf[pos:])
            # b'\x00000001\x95\n\x87\x0e\x01\x01\x05\x00\xb1\xb9\xd6\r\xc7\x0e\x8d\xd7\x1a\x84\x04S\x9c<M\xb6\xc8\x0e\x97\x8e\x0c\x00\xae\n\x00\x01\xa0\x1e\x9e\xb3\x03A\x02\x84\xf9\x01\xa8|B\x03\x8c\xd6\x01\xb0lC\x04\xb7\xdb\x02\xac\x7fD\x05\xbb\xb0\x01\xbe\xa0\x01y\x08\x01GC\x04\x00\x00\x95\n'
            market, code, active1 = struct.unpack("<B6sH", body_buf[pos: pos + 9])
            pos += 9

            price, pos = get_price(body_buf, pos)  # noqa
            last_close_diff, pos = get_price(body_buf, pos)

            open_diff, pos = get_price(body_buf, pos)
            high_diff, pos = get_price(body_buf, pos)
            low_diff, pos = get_price(body_buf, pos)

            reversed_bytes0, pos = get_price(body_buf, pos)
            reversed_bytes1, pos = get_price(body_buf, pos)

            vol, pos = get_price(body_buf, pos)
            cur_vol, pos = get_price(body_buf, pos)

            amount_raw, = struct.unpack("<I", body_buf[pos: pos + 4])
            amount = get_volume(amount_raw)
            pos += 4

            s_vol, pos = get_price(body_buf, pos)  # noqa
            b_vol, pos = get_price(body_buf, pos)

            reversed_bytes2, pos = get_price(body_buf, pos)
            reversed_bytes3, pos = get_price(body_buf, pos)

            bid1, pos = get_price(body_buf, pos)
            ask1, pos = get_price(body_buf, pos)

            bid_vol1, pos = get_price(body_buf, pos)
            ask_vol1, pos = get_price(body_buf, pos)

            bid2, pos = get_price(body_buf, pos)
            ask2, pos = get_price(body_buf, pos)

            bid_vol2, pos = get_price(body_buf, pos)
            ask_vol2, pos = get_price(body_buf, pos)

            bid3, pos = get_price(body_buf, pos)  # noqa
            ask3, pos = get_price(body_buf, pos)

            bid_vol3, pos = get_price(body_buf, pos)
            ask_vol3, pos = get_price(body_buf, pos)

            bid4, pos = get_price(body_buf, pos)
            ask4, pos = get_price(body_buf, pos)

            bid_vol4, pos = get_price(body_buf, pos)
            ask_vol4, pos = get_price(body_buf, pos)

            bid5, pos = get_price(body_buf, pos)
            ask5, pos = get_price(body_buf, pos)

            bid_vol5, pos = get_price(body_buf, pos)
            ask_vol5, pos = get_price(body_buf, pos)

            reversed_bytes4 = struct.unpack("<H", body_buf[pos: pos + 2])
            pos += 2

            reversed_bytes5, pos = get_price(body_buf, pos)
            reversed_bytes6, pos = get_price(body_buf, pos)

            reversed_bytes7, pos = get_price(body_buf, pos)
            reversed_bytes8, pos = get_price(body_buf, pos)

            reversed_bytes9, active2 = struct.unpack("<hH", body_buf[pos: pos + 4])
            pos += 4

            code = code.decode("utf-8")
            coefficient = get_security_coefficient(market, code)[0]

            one_stock = OrderedDict(
                [
                    ("market", market),
                    ("code", code),
                    ("active1", active1),
                    ("price", self._cal_price(price, 0, coefficient)),
                    ("last_close", self._cal_price(price, last_close_diff, coefficient)),
                    ("open", self._cal_price(price, open_diff, coefficient)),
                    ("high", self._cal_price(price, high_diff, coefficient)),
                    ("low", self._cal_price(price, low_diff, coefficient)),
                    ("servertime", self._format_time("%s" % reversed_bytes0)),
                    ("reversed_bytes0", reversed_bytes0),
                    ("reversed_bytes1", reversed_bytes1),
                    ("vol", vol),
                    ("cur_vol", cur_vol),
                    ("amount", amount),
                    ("s_vol", s_vol),
                    ("b_vol", b_vol),
                    ("reversed_bytes2", reversed_bytes2),
                    ("reversed_bytes3", reversed_bytes3),
                    ("bid1", self._cal_price(price, bid1, coefficient)),
                    ("ask1", self._cal_price(price, ask1, coefficient)),
                    ("bid_vol1", bid_vol1),
                    ("ask_vol1", ask_vol1),
                    ("bid2", self._cal_price(price, bid2, coefficient)),
                    ("ask2", self._cal_price(price, ask2, coefficient)),
                    ("bid_vol2", bid_vol2),
                    ("ask_vol2", ask_vol2),
                    ("bid3", self._cal_price(price, bid3, coefficient)),
                    ("ask3", self._cal_price(price, ask3, coefficient)),
                    ("bid_vol3", bid_vol3),
                    ("ask_vol3", ask_vol3),
                    ("bid4", self._cal_price(price, bid4, coefficient)),
                    ("ask4", self._cal_price(price, ask4, coefficient)),
                    ("bid_vol4", bid_vol4),
                    ("ask_vol4", ask_vol4),
                    ("bid5", self._cal_price(price, bid5, coefficient)),
                    ("ask5", self._cal_price(price, ask5, coefficient)),
                    ("bid_vol5", bid_vol5),
                    ("ask_vol5", ask_vol5),
                    ("reversed_bytes4", reversed_bytes4),
                    ("reversed_bytes5", reversed_bytes5),
                    ("reversed_bytes6", reversed_bytes6),
                    ("reversed_bytes7", reversed_bytes7),
                    ("reversed_bytes8", reversed_bytes8),
                    ("reversed_bytes9", reversed_bytes9 / 100.0),  # 涨速
                    ("active2", active2),
                ]
            )

            stocks.append(one_stock)

        return stocks

    @staticmethod
    def _cal_price(base_p, diff, coefficient=0.01):
        """

        :param base_p:
        :param diff:
        :return:
        """
        return float(base_p + diff) * coefficient

    @staticmethod
    def _format_time(time_stamp):
        """
        format time from reversed_bytes0
        by using method from https://github.com/rainx/pytdx/issues/187
        :param time_stamp:
        :return:
        """
        time = time_stamp[:-6] + ":"

        if int(time_stamp[-6:-4]) < 60:
            time += "%s:" % time_stamp[-6:-4]
            time += "%06.3f" % (int(time_stamp[-4:]) * 60 / 10000.0)
        else:
            time += "%02d:" % (int(time_stamp[-6:]) * 60 / 1000000)
            time += "%06.3f" % ((int(time_stamp[-6:]) * 60 % 1000000) * 60 / 1000000.0)

        return time



# cython: language_level=3

import struct


class TdxFileNotFoundException(Exception):
    pass


class TdxNotAssignVipdocPathException(Exception):
    pass


class BaseReader(object):
    def __init__(self, vipdoc_path=None):
        """
        构造函数
        :param vipdoc_path: tdx 的 vipdoc 目录
        """
        self.vipdoc_path = vipdoc_path

    @staticmethod
    def unpack_records(fmt, data):
        """
        解包
        :param fmt:
        :param data:
        :return:
        """
        record = struct.Struct(fmt)
        result = (record.unpack_from(data, offset) for offset in range(0, len(data), record.size))

        return result

    def get_df(self, code_or_file, exchange=None):
        """
        转换 pd.DataFrame
        :param code_or_file:
        :param exchange:
        """
        raise NotImplementedError("not yet")

    @staticmethod
    def _parse_date(num):
        """
        解析日期
        :param num:
        :return:
        """
        month = (num % 2048) // 100
        year = num // 2048 + 2004
        day = (num % 2048) % 100

        return year, month, day

    @staticmethod
    def _parse_time(num):
        """
        解析时间
        :param num:
        :return:
        """
        return (num // 60), (num % 60)



# cython: language_level=3

import struct
from collections import OrderedDict

from tdxpy.helper import get_price
from tdxpy.helper import get_time
from tdxpy.parser.base import BaseParser


class GetTransactionData(BaseParser):
    def setParams(self, market, code, start, count):
        """
        设置请求参数
        :param market:
        :param code:
        :param start:
        :param count:
        """
        if type(code) is str:
            code = code.encode("utf-8")

        pkg = bytearray.fromhex("0c 17 08 01 01 01 0e 00 0e 00 c5 0f")
        pkg.extend(struct.pack("<H6sHH", market, code, start, count))

        self.send_pkg = pkg

    def parseResponse(self, body_buf):
        """
        解析返回结果
        :param body_buf:
        :return:
        """
        pos = 0

        num, = struct.unpack("<H", body_buf[:2])

        pos += 2

        ticks = []
        last_price = 0

        for _ in range(num):
            hour, minute, pos = get_time(body_buf, pos)
            price_raw, pos = get_price(body_buf, pos)

            vol, pos = get_price(body_buf, pos)
            num, pos = get_price(body_buf, pos)

            buy_or_sell, pos = get_price(body_buf, pos)
            _, pos = get_price(body_buf, pos)

            last_price = last_price + price_raw

            tick = OrderedDict(
                [
                    ("time", "%02d:%02d" % (hour, minute)),
                    ("price", float(last_price) / 100),
                    ("vol", vol),
                    ("num", num),
                    ("buyorsell", buy_or_sell),  # noqa
                ]
            )

            ticks.append(tick)

        return ticks



# cython: language_level=3

import logging

formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
console = logging.StreamHandler()
console.setLevel(logging.DEBUG)
console.setFormatter(formatter)

logger = logging.getLogger("tdxpy")
logger.addHandler(console)
logger.setLevel(logging.INFO)



# cython: language_level=3

import datetime
import functools
import socket
import threading
import time

import pandas as pd

from tdxpy.exceptions import TdxConnectionError
from tdxpy.exceptions import TdxFunctionCallError
from tdxpy.exceptions import ValidationException
from tdxpy.heartbeat import HeartBeatThread
from tdxpy.logger import logger
from tdxpy.parser.raw_parser import RawParser

DEFAULT_HEARTBEAT_INTERVAL = 10.0

CONNECT_TIMEOUT = 5.000
RECV_HEADER_LEN = 0x10


def last_ack_time(func):
    """
    装饰器: 更新最后 ack 时间
    :param func:
    :return:
    """

    @functools.wraps(func)
    def wrapper(self, *args, **kw):
        self.last_ack_time = time.time()

        logger.debug(f"last ack time update to {self.last_ack_time}")

        ret = None

        try:
            ret = func(self, *args, **kw)
        except (TypeError, ValueError) as e:
            raise ValidationException(*e.args)
        except Exception as e:
            # logger.exception(e)
            current_exception = e
            logger.debug(f"hit exception on req exception is {e}")

            if self.auto_retry and self.client:
                for time_interval in self.retry_strategy.generate():
                    try:
                        time.sleep(time_interval)

                        self.disconnect()
                        self.connect(self.ip, self.port)

                        ret = func(self, *args, **kw)
                        return ret

                    except Exception as retry_e:
                        logger.debug(f"hit exception on *retry* req exception is {retry_e}")
                        current_exception = retry_e

                logger.debug("perform auto retry on req ")

            # 最后交易失败
            # self last_transaction_failed is True

            # 如果 raise_exception=True 抛出异常
            if self.raise_exception:
                to_raise = TdxFunctionCallError("calling function error")
                to_raise.original_exception = current_exception
                raise to_raise

        return ret

    return wrapper


class RetryStrategy(object):
    @classmethod
    def generate(cls):
        raise NotImplementedError("need to override")


class DefaultRetryStrategy(RetryStrategy):
    """
    默认的重试策略，您可以通过写自己的重试策略替代本策略, 改策略主要实现 generate 方法，该方法是一个生成器，
    返回下次重试的间隔时间, 单位为秒，我们会使用 time.sleep在这里同步等待之后进行重新connect,然后再重新发起
    源请求，直到 generate 结束。
    """

    @classmethod
    def generate(cls):
        for time_interval in [0.1, 0.5, 1, 2]:
            yield time_interval


class TrafficStatSocket(socket.socket):
    """
    实现支持流量统计的socket类
    """

    def __init__(self, sock, mode):
        super(TrafficStatSocket, self).__init__(sock, mode)

        # 流量统计相关
        self.send_pkg_num = 0  # 发送次数
        self.recv_pkg_num = 0  # 接收次数
        self.send_pkg_bytes = 0  # 发送字节
        self.recv_pkg_bytes = 0  # 接收字节数
        self.first_pkg_send_time = None  # 第一个数据包发送时间

        self.last_api_send_bytes = 0  # 最近的一次api调用的发送字节数
        self.last_api_recv_bytes = 0  # 最近一次api调用的接收字节数


class BaseSocketClient(object):
    def __init__(self, multithread=False, heartbeat=False, auto_retry=False, raise_exception=False):
        """
        构造函数
        :param multithread: 是否多线程
        :param heartbeat: 是否心跳
        :param auto_retry: 是否自动回复
        :param raise_exception: 是否抛出异常
        """

        self.need_setup = True
        self.closed = True
        self.client = None

        self.lock = (multithread or heartbeat) and threading.Lock() or None

        self.heartbeat = heartbeat
        self.heartbeat_thread = None

        self.stop_event = None
        self.heartbeat_interval = DEFAULT_HEARTBEAT_INTERVAL  # 默认10秒一个心跳包

        self.last_ack_time = time.time()
        # self last_transaction_failed = False

        self.ip = None
        self.port = None

        # 是否重试
        self.auto_retry = auto_retry

        # 可以覆盖这个属性，使用新的重试策略
        self.retry_strategy = DefaultRetryStrategy()

        # 是否在函数调用出错的时候抛出异常
        self.raise_exception = raise_exception

    def connect(self, ip: str = None, port: int = 7709, time_out=CONNECT_TIMEOUT, bind_port=None, bind_ip="0.0.0.0"):
        """
        连接服务器

        :param ip:  服务器ip 地址
        :param port:  服务器端口
        :param time_out: 连接超时时间
        :param bind_port: 绑定的本地端口
        :param bind_ip: 绑定的本地ip
        :return: 是否连接成功 True/False
        """

        if not ip:
            raise ValidationException('IP Address bad.')

        self.client = TrafficStatSocket(socket.AF_INET, socket.SOCK_STREAM)
        self.client.settimeout(time_out)

        logger.debug(f"connecting to server : {ip} on port :{port:d}")

        try:
            self.ip = ip
            self.port = port

            if bind_port:
                self.client.bind((bind_ip, bind_port))

            self.client.connect((ip, port))
        except socket.timeout as e:
            logger.exception(e)
            logger.debug("connection expired")

            if self.raise_exception:
                raise TdxConnectionError("connection timeout error")

            return False
        except Exception:  # noqa
            if self.raise_exception:
                raise TdxConnectionError("other errors")

            return False

        self.closed = False
        logger.debug("connected!")

        if self.need_setup:
            self.setup()

        # 启动心跳包
        if self.heartbeat:
            self.stop_event = threading.Event()
            self.heartbeat_thread = HeartBeatThread(client=self, stop_event=self.stop_event, interval=self.heartbeat_interval)
            self.heartbeat_thread.start()

        return self

    def disconnect(self):
        """
        断开连接
        """

        # 停止心跳事件
        if self.heartbeat_thread and self.heartbeat_thread.is_alive():
            self.stop_event.set()

        if self.client:
            logger.debug("disconnecting")

            try:
                self.client.shutdown(socket.SHUT_RDWR)
                self.client.close()
                self.client = None
            except Exception as e:
                logger.debug(f'disconnect err: {e}')
                if self.raise_exception:
                    raise TdxConnectionError(f"disconnect err: {e}")
            finally:
                self.client = None

            logger.debug("disconnected")

    def close(self):
        """
        disconnect的别名，为了支持 with closing(obj): 语法
        :return:
        """
        self.disconnect()

    def get_traffic_stats(self):
        """
        获取流量统计的信息
        :return:
        """
        if self.client.first_pkg_send_time is not None:
            total_seconds = (datetime.datetime.now() - self.client.first_pkg_send_time).total_seconds()

            if total_seconds != 0:
                send_bytes_per_second = self.client.send_pkg_bytes // total_seconds
                recv_bytes_per_second = self.client.recv_pkg_bytes // total_seconds
            else:
                send_bytes_per_second = None
                recv_bytes_per_second = None
        else:
            total_seconds = None
            send_bytes_per_second = None
            recv_bytes_per_second = None

        return {
            "send_pkg_num": self.client.send_pkg_num,
            "recv_pkg_num": self.client.recv_pkg_num,
            "send_pkg_bytes": self.client.send_pkg_bytes,
            "recv_pkg_bytes": self.client.recv_pkg_bytes,
            "first_pkg_send_time": self.client.first_pkg_send_time,
            "total_seconds": total_seconds,
            "send_bytes_per_second": send_bytes_per_second,
            "recv_bytes_per_second": recv_bytes_per_second,
            "last_api_send_bytes": self.client.last_api_send_bytes,
            "last_api_recv_bytes": self.client.last_api_recv_bytes,
        }

    def send_raw_pkg(self, pkg):
        """
        发送原始数据包
        # for debuging and testing protocol

        :param pkg:
        :return:
        """
        cmd = RawParser(self.client, lock=self.lock)
        cmd.setParams(pkg)

        return cmd.call_api()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    @staticmethod
    def to_df(v):
        """
        数据转换 df 格式
        :param v:
        :return:
        """

        v = v or ''

        if not v:
            return pd.DataFrame(data=None)

        if isinstance(v, list):
            return pd.DataFrame(data=v)

        if isinstance(v, dict):
            return pd.DataFrame(data=[v])

        return pd.DataFrame(data=[{"value": v}])

    def setup(self):
        """
        方法
        """
        pass



# cython: language_level=3

from tdxpy.parser.base import BaseParser


class RawParser(BaseParser):
    def setParams(self, pkg):
        """
        设置发送数据包
        :param pkg: send pkg
        """
        self.send_pkg = pkg

    def parseResponse(self, body_buf):
        """
        解析结果
        :param body_buf: buff
        :return:
        """
        return body_buf
